#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# checked with 3.10
# Created by H. Inoue on 2024-02-17
#########################################################################################
# csvファイルをPPP共通仕様協議会が定めるNGSI V2の規格に合致するJSON文に変換します
# Input   inputfile: 入力となるcsvファイル。UTF-8。FMシステムの出力を想定しています
#         schemafile: csvからJSONに変換する際のattributeと項目との対応を定義するファイル。UTF-8
#                   形式はjson schemaを拡張した形式です
#         dir: 各種データファイルが存在するディレクトリのパス
#         addressfile: アドレスのファイルのファイル名。既定値は'mt_city_all.csv'
# Output outputfile:出力するcsvファイル。UTF-8。schemaに指定した項目順に項目は並ぶ。csvファイルに
#         対応する項目値が無い場合は、Attribute自体を省略して出力する。
# オプション
#         inputdir:  inputfileが存在するティレクトリのバス。指定しないと'dir'の指定が採用される
#         outputdir: outputfileが存在するティレクトリのバス。指定しないと'dir'の指定が採用される
#         schemadir: schemafileが存在するティレクトリのバス。指定しないと'dir'の指定が採用される
#         <Entity type>file: データを変換する際に参照するEntityのリスト
#         <Entity type>dir: <Entity type>fileが存在するティレクトリのバス。指定しないと'dir'の指定が採用される
#         addressdir: schemafileが存在するティレクトリのバス。指定しないと'dir'の指定が採用される
#         jsontype: outputfileのJSON文の形式。JSONとしては以下の3種類をサポートする
#                      jsontype=entity : 波括弧に囲まれるひとつのEntity。最初の一件しか出力されない
#                      jsontype=array : 角かっこに囲まれるEntityの列。<Entity type>fileで指定する
#                                       ファイルはこの形式
#                      jsontype=batch : op/updateに対し指定する、一括登録のためのJSON文(既定値)

import json
import csv
import sys
from my_package.NGSIv2 import GET_entities, exschema, entity, trace_log, error_log
from my_package.address import address
from my_package.CoppellTechnologies_common import CT_text, CT_csv
#########################################################################################
# 各変数の初期化
def initialize_run():
    run={
        'param': {
            'dir': 'c:/Tools/data/',
            'inputdir': '',
            'inputfile': '',
            'outputdir': '',
            'outputfile': '',
            'outputtype': '',
            'outputmode': 'extend',
            'schemadir': '',
            'schemafile': '',
            'Organizationdir': '',
            'Organizationfile': '',
            'Departmentdir': '',
            'Departmentfile': '',
            'Landdir': '',
            'Landfile': '',
            'Buildingdir': '',
            'Buildingfile': '',
            'Facilitydir': '',
            'Facilityfile': '',
            'FacilityStatisticsdir': '',
            'FacilityStatisticsfile': '',
            'BuildingComponentdir': '',
            'BuildingComponentfile': '',
            'Devicedir': '',
            'Devicefile': '',
            'DeviceModeldir': '',
            'DeviceModelfile': '',
            'Complaintdir': '',
            'Complaintfile': '',
            'Reportdir': '',
            'Reportfile': '',
            'addressdir': 'c:/Tools/data/',
            'addressfile': 'mt_city_all.csv',
            'logdir': '',
            'logfile': 'ctojlog.txt',
            'jsontype': 'array',
            'format': 'normalized',
            'truevalue': ['True', 'true', '真', '有'],
            'falsevalue': ['False', 'false', '偽', '無']
        },
        'handles': {},
        'mode': 'debug',
        'errorflag': False,
        'onerror': 'continue',
        'types': ['Text','Number','Integer','Boolean','Date', 'DateTime','URL','Time','Relationship'],
        'entitynames': ['Organization','Department','Land','Building',
            'Facility','FacilityStatistics','BuildingComponent','Device',
            'DeviceModel','Complaint','Report'],
        'errorobj': {},
        'errorinfo': {'attr': ''},
        'warning': False,
        'type_tbl': {'str':'string', 'int':'integer','float':'number', 'bool':'boolean' }
    }
    param = run['param']
    if len(sys.argv) > 1:
        for arg in sys.argv[1:]:
            p = arg.split('=')
            if len(p) != 2:
                input(f'## error ## 不正なパラメータです: {arg}')
                exit()
            if p[0] in param.keys():
                param[p[0]] = p[1]
            else:
                input( f'## error ## 未知のパラメータです: {p[0]}')
                exit()
    if (param['inputdir'] == ''):
        param['inputdir'] = param['dir']
    if (param['outputdir'] == ''):
        param['outputdir'] = param['dir']
    if (param['schemadir'] == ''):
        param['schemadir'] = param['dir']
    if (param['outputdir'] == ''):
        param['outputdir'] = param['dir']
    if (param['addressdir'] == ''):
        param['addressdir'] = param['dir']
    for a1 in run['entitynames']:
        if ((a1 + 'dir') in param and param[a1 + 'dir'] == ''):
            param[a1 + 'dir'] = param['dir']
    if (param['logdir'] == ''):
        param['logdir'] = param['dir']
    return run

def put_entity(run: dict,
               head_row: list,
               data_row: list)->dict:
    """
    拡張Json schemaの指定に従い、csvの一行に対応するlistからEntity一つ分のdictを作成します
    run: ツール内の共通テーブルです
    head_row: csvの最初の一行に対応するlistです。項目名のlistになっています
    data_row: csvの二行目以降の一行に対応するlistです。項目名に対応するデータのlistです
    返却値: 作成したEntityに対応するdictです
    """
    schema = run['scm']['properties']          # 最上位のtypeは必ずobjectなので決め打ち
    err = run['handles']['error']
    entity={}
    flags = {'skipped': False, 'settled': False}  # Entityレベルでは不要だが、I/Fを合わせる
    for attr_name,attr_value in schema.items():# schemaのattribute毎にループ
        run['errorinfo']['attr'] = attr_name
        attr_type = attr_value['type']
        if attr_type == 'object':              # normalized形式なので、idとtype以外はobject
            d1 = process_object(run,attr_value, head_row, data_row, flags)
            if d1 == {}:
                continue
        elif attr_type == 'array':             # AttributeのタイプはArrayです。normalized formatではあり得ない
            err.set_errorobj('01', f'## schema ## attribute直下にarrayが現れた. #attr#')
        else:                                  # idとtypeの場合だけ。string、number等の「値」かconst
            # print(f'## process_entity(value): schema: ',attr_value)
            dataSource = attr_value.get('dataSource',{})
            if not dataSource:
                continue
            d1 = process_value(run,dataSource, head_row, data_row, flags)
            # print(f'## put_entity(value): value: ',d1)
        if d1 == '':                           # Attributeに設定すべき値が無かった
            continue                           # Attributeは作成しない
        entity[attr_name] = d1                 # attributeを辞書形式の要素(k-vペア)で作成します
    if entity.get('id') == None or entity.get('type') == None:
        # print(f'put_entity(error02): entity: ',entity)
        err.set_errorobj('02', f'idまたはtypeの生成に失敗しています')
    return(entity)                             # Entity一つ分の辞書を返却します

def process_object(run: dict,
                   schema: dict,
                   head_row: list,
                   data_row: list,
                   flags: dict)-> dict:    # Attributeの値の辞書を返却します
    """
    拡張json schemaの中でtypeがobjectの場合の処理です。
    schema: typeがobjectの拡張json schemaのdictです。このdictには必ずpropertiesが存在
            します。
    head_row: csvの最初の一行に対応するlistです。項目名のlistになっています
    data_row: csvの二行目以降の一行に対応するlistです。項目名に対応するデータのlistです
    返却値: Attributeの値です。形式はstr, int, boolean, dict, listです。
            csvファイルに値が存在しなかった場合は{}を返却します
    """
    handles =run['handles']
    err = handles['error']
    # attributeのvalueのタイプにより処理を切り分けます
    # print(f'## process_object(top). schema: ',schema)
    prop = schema.get('properties','')
    required = schema.get('required','')
    flags_sub = {'skipped': False, 'settled': False}
    ret_value = {}
    for attr_key,attr_value in prop.items():   # properties内の各sub-attributeについてループ
        attr_type = attr_value.get('type','')
        if attr_type == 'object':
            d1 = process_object(run,attr_value, head_row, data_row, flags_sub)
            if d1 == {}:
                continue
            ret_value[attr_key] = d1
        elif attr_type == 'array':             # AttributeのタイプはArrayです
            attr_items = attr_value.get('items','')# typeがarrayの時は必ずitemsの指定が必要です
            if isinstance(attr_items,list):    # itemsがリストの場合
                ret_value_sub = []
                lost_value = False             # 順序付きリストで、欠測があったことを示すフラグ
                for v2 in attr_items:          # itemsのリストの各要素に対し処理します
                    if v2['type'] == 'array':  # この場合、必ず順序付きリストです
                        r0 = process_array(run, v2, head_row, data_row, flags_sub)
                        if r0 == []:
                            lost_value = True
                        elif lost_value == True:   # 欠測後に無視する値があった
                                handles['trace_log'].set_log(2,'01',f'順序付きリストの途中の値が不足しているため出力できない項目がありました. attribute name: {attr_key}')
                                flags_sub['skipped'] = True
                                flags_sub['settled'] = False
                                continue
                        else:
                            ret_value_sub.append(r0)
                    elif v2['type'] == 'object':   # オブジェクトのリスト。欠測があっても構わない
                        r0 = process_object(run, v2, head_row, data_row, flags_sub)
                        if r0 != {}:
                            ret_value_sub.append(r0)
                    else:                      # stringやnumber等の値
                        dataSource = v2.get('dataSource', {})
                        r0 = process_value(run, dataSource, head_row, data_row, flags_sub)
                        if r0 == '':
                            lost_value = True
                        elif lost_value == True:
                            handles['trace_log'].set_log(2,'03',f'順序付きリストの途中の値が不足しているため出力できない項目がありました. attribute name: {attr_key}')
                            flags_sub['skipped'] = True
                            flags_sub['settled'] = False
                        else:
                            ret_value_sub.append(r0)
                ret_value[attr_key] = ret_value_sub
            else:                              # itemsの形式がlistではない
                ret_value_sub = process_array(run, attr_value, head_row, data_row, flags_sub)
                if ret_value_sub == []:
                    continue
                ret_value[attr_key] = ret_value_sub
        else:                                  # typeがstring, number等の「値」
            dataSource = attr_value.get('dataSource', {})
            # print(f'process_object(dataSource): dataSource: ', dataSource)
            if not dataSource:
                continue
            ret_value_sub = process_value(run,dataSource, head_row, data_row, flags_sub)
            # print(f'process_object(return): return: ', ret_value_sub)
            if ret_value_sub == '':
                continue
            ret_value[attr_key] = ret_value_sub
    if flags_sub['settled']:                   # const指定以外で値が設定された
        flags['settled'] = True
    elif flags_sub['skipped']:                 # 値の設定がスキップされた項目があった
        flags['skipped'] = True
        return {}
    format = schema.get('format', '')
    if format:
        format_object(run, ret_value, format)
    return ret_value

def process_array(run,
                  schema: dict,
                  head_row: list,
                  data_row: list,
                  flags: dict)-> list:
    """
    拡張json schemaの中でtypeがarrayの場合の処理です。
    schema: typeがarrayの拡張json schemaのdictです。このdictには必ずitemsが存在
            します。
    head_row: csvの最初の一行に対応するlistです。項目名のlistになっています
    data_row: csvの二行目以降の一行に対応するlistです。項目名に対応するデータのlistです
    返却値: Attributeの値です。形式はstr, int, boolean, dict, listです。
            csvファイルに値が存在しなかった場合は[]を返却します
    """
    handles =run['handles']
    items = schema['items']
    # print(f'## process_array(top). schema: ',schema)
    ret_list = []
    if isinstance(items,list):
        for items_sub in items:
            items_type = items_sub['type']
            if items_type == 'array':                  # arrayの要素がarray
                r1 = process_array(run,items_sub, head_row, data_row, flags)
                if r1 != []:
                    ret_list += [r1]
            elif items_type == 'object':               # arrayの要素がsub-attribute
                ret_obj = process_object(run, items_sub, head_row, data_row, flags)
                if ret_obj != {}:
                    ret_list += [ret_obj]
            else:
                dataSource = items_sub.get('dataSource',{})
                ret_val = process_value (run, dataSource, head_row, data_row, flags)
                if ret_val != '':
                    ret_list += [ret_val]
    else:
        items_type = items['type']
        if items_type == 'array':                  # arrayの要素がarray
            r1 = process_array(run,items, head_row, data_row, flags)
            # print(f'## process_array(nested array). value: {str(r1)}, items: ',items)
            if r1 != []:
                ret_list = [r1]
        elif items_type == 'object':               # arrayの要素がsub-attribute
            ret_obj = process_object(run, items, head_row, data_row, flags)
            # print(f'## process_array(object). schema: {str(schema)}, return: {str(ret_obj)}')
            if ret_obj != {}:
                ret_list = [ret_obj]
        else:
            dataSource = items.get('dataSource',{})
            ret_val = process_value (run, dataSource, head_row, data_row, flags)
            if ret_val != '':
                ret_list = [ret_val]
    return ret_list

def process_value (run, schema, head_row, data_row, flags):
    """
    拡張json schemaの中でtypeがobjectでもarrayでもない場合の処理です。
    schema: 拡張json schemaのdictです。このdictには必ず値を求める記述が存在
            します。
    head_row: csvの最初の一行に対応するlistです。項目名のlistになっています
    data_row: csvの二行目以降の一行に対応するlistです。項目名に対応するデータのlistです
    返却値: Attributeの値です。形式はstr, int, booleanです。
            csvファイルに値が存在しなかった場合は''を返却します
    """
    handles = run['handles']
    # print(f'## process_value(top): schema', schema)
    # 最初にdataSourceがlistなのか値なのかを調べます。
    if isinstance(schema, list):
        # dataSourceがlistで指定された場合、
        ret_value = ''
        for schema_sub in schema:
            ret_value_sub = process_value (run, schema_sub, head_row, data_row, flags) 
            ret_value += ret_value_sub 
        return ret_value
    # まずは最優先のentityTypeで指定したEntity typeから求めます
    if schema.get('entityType'):           # entityTypeの指定がある
        entityType = schema['entityType']
        # print(f'process_value(entities). entityType: {entityType}')
        if entityType not in run['entitynames']:
            handles['error'].set_errorobj( '01', f'## schema ## unknown Entity type : {entityType}')
        if handles.get(entityType) == None:
            handles['error'].set_errorobj( '02', f'## error ## entityTypeに指定したEntityがパラメータで指定されていない. Entity type : {entityType}')
        valueAttr = schema['valueAttr']
        if valueAttr == '':
                handles['error'].set_errorobj( '02', f'## schema ## valueAttrの指定が無い. entityType: {entityType}')
        searchAttr = schema['searchAttr']
        if searchAttr == '':
                handles['error'].set_errorobj( '03', f'## schema ## searchAttrの指定が無い. entityType: {entityType}')
        cvalue = process_value (run, schema['dataSource'], head_row, data_row, flags)
        if not cvalue:
            # print(f'## process_value(skip). dataSource: ', schema['dataSource'])
            flags['skipped'] = True
            return ''
        if searchAttr == 'identificationGroup': # 特別な処理が必要
            identificationType = schema.get('identificationType','')
            if identificationType == '':       # identificationTypeの指定も同時に必要
                handles['error'].set_errorobj( '04', f'## schema ## searchAttrがidentificationGroupの場合はidentificationTypeの指定は必須です.  entityType: {entityType}')
            ent = handles[entityType].get_listed_subattr ('identificationGroup', 'identificationType', identificationType, 'identification', cvalue)
        else:
            p = searchAttr.split('.')
            # print(f'## process_value(get): entity type: {entityType}, searchAttr: {searchAttr}')
            if len(p) == 1:
                ent = handles[entityType].get(searchAttr, cvalue)
            elif len(p) == 2:
                # print(f'## process_value(subattr): handles: ', handles)
                ent = handles[entityType].get_subattr(p[0], p[1], cvalue)
            else:
                handles['error'].set_errorobj( '05', f'## schema ## dataSourceに記述できるピリオドは最大1つです. entityType: {entityType}, dataSource: {searchAttr}')

        if ent== {}:                           # 一致するEntityが無かった
            # msg = str(handles[entityType].entities)[:200]
            # print(f'## prcess_value. entityType: {entityType}, key: {searchAttr}, entities: {msg}')
            handles['trace_log'].set_log(2,'01',f'entityTypeで指定したEntity typeに検索値が一致するEntityが見つかりません. entityType: {entityType}, key: {searchAttr}, 検索値: {cvalue}')
            ret_value = ''
            flags['skipped'] = True
            return ''
        
        p0 = valueAttr.split('.')
        p1 = valueAttr.split('[')        # listの要素指定
        # print(f'## process_value(array-1). valueAttr: {str(p1)}, length: {len(p1)}')
        for i in range(1,len(p1)):
            if len(p1[i])<2 or p1[i][-1:] != ']':
                handles['error'].set_errorobj('06', f'## schema ## valueAttrのインデックス指定の形式が正しくない.  valueAttr: {valueAttr}')
            p1[i] = int(p1[i][:-1])
            # print(f'## process_value(array-2). index: {i}, index value: {p1[i]}')
        if len(p0) == 1:
            if valueAttr in ('id', 'type'):
                ret_value = schema.get('prefix', '')+str(ent[valueAttr])
                flags['settled'] = True
            elif p1[0] not in ent:
                handles['trace_log'].set_log(2,'02',f'searchAttrで検索したEntity内にvalueAttrで指定したAttributeがありません. Entity id: '+ent['id']+f', valueAttr: {valueAttr}')
                ret_value = ''
                flags['skipped'] = True
            else:
                if len(p1) == 1:
                    ret_value = schema.get('prefix', '')+str(ent[valueAttr]['value'])
                elif len(p1) == 2:
                    v0 = ent[p1[0]]['value'][p1[1]] if p1[1] < len(ent[p1[0]]['value']) else ''
                    ret_value = schema.get('prefix', '')+str(v0) if v0 else ''
                elif len(p1) == 3:
                    v0 = ent[p1[0]]['value'][p1[1]][p1[2]] if p1[1] < len(ent[p1[0]]['value']) and p1[2] < len(ent[p1[0]]['value'][p1[1]]) else ''
                    ret_value = schema.get('prefix', '')+str(v0) if v0 else ''
                else:
                    handles['error'].set_errorobj('04', f'## schema ## valueAttrに指定できる次元は2次元までです. valueAttr {valueAttr}')
                flags['settled'] = True
        elif len(p0) == 2:
            if p0[0] == 'identificationGroup':    # 特別な処理が必要
                if p0[1] != 'identification':     # 項目はidentificationしかあり得ない
                    handles['error'].set_errorobj('04', f'## schema ## valueAttrがidentificationGroupの場合、sub-attributeはidentificationです.  entityType: {entityType}')
                identificationType = schema.get('identificationType','')
                if identificationType == '':       # identificationTypeの指定も同時に必要
                    handles['error'].set_errorobj('05', f'## schema ## valueAttrがidentificationGroupの場合はidentificationTypeの指定は必須です.  entityType: {entityType}')
                ret_value = ''
                for ig in ent['identificationGroup']['value']:
                    if ig['identificationType'] == identificationType:
                        ret_value = schema.get('prefix', '')+ig['identification']  # idの値を返却
                        break
                if ret_value == '':
                    handles['trace_log'].set_log(2,'03',f'identificationTypeの指定に一致する要素が見つかりません. Entity id: '+ent['id'])
                    flags['skipped'] = True
                else:
                    flags['settled'] = True
            else:
                if p0[0] not in ent:
                    handles['trace_log'].set_log(2,'04',f'searchAttrで検索したEntity内にvalueAttrで指定したAttributeがありません. Entity id: '+ent['id']+f', valueAttr: {valueAttr}')
                    ret_value = ''
                    flags['skipped'] = True
                elif p0[1] not in ent[p[0]]['value']:
                    handles['trace_log'].set_log(2,'05',f'searchAttrで検索したEntity内にvalueAttrで指定したSub-attributeがありません. Entity id: '+ent['id']+f', valueAttr: {valueAttr}')
                    ret_value = ''
                    flags['skipped'] = True
                else:
                    ret_value = schema.get('prefix', '')+str(ent[p[0]]['value'][p[1]]) # MUST normalized format
                    flags['settled'] = True
        else:
            handles['error'].set_errorobj( '05', f'## schema ## searchAttrに記述できるピリオドは最大1つです. entityType: {entityType}, dataSource: {searchAttr}')
        return ret_value
    
    # entityTypeの指定が無いときは、csvから求めます
    cvalue = ret_csv_value (run, schema, head_row, data_row)
    # print(f'## process_value(cvalue): cvalue: ',cvalue)
    if cvalue != '':
        if schema['type'] == 'string':
            ret_value = schema.get('prefix', '')+cvalue
        else:
            ret_value = cvalue
        flags['settled'] = True
        return ret_value
    # csvから値を求められなかったので、alernateの指定から求めます
    alter_map = schema.get('alternate','')     # alretnateで指定した代替mapを探します
    if alter_map != '':
        # 最初にdalternateがlistなのか値なのかを調べます。
        if isinstance(alter_map, list):
            # dataSourceがlistで指定された場合、
            ret_value = ''
            for alter_map_sub in alter_map:
                ret_value_sub = process_value (run, alter_map_sub, head_row, data_row, flags) 
                ret_value += ret_value_sub 
            return ret_value
        
        ret_value = process_value (run, alter_map, head_row, data_row, flags) 
        return ret_value
    
    # 結局値を求められなかったので、既定値を返却します
    if schema.get('const') != None:
        return (schema.get('const'))
    flags['skipped'] = True
    return ''

def ret_csv_value (run,schema, head_row, data_row):
    handles =run['handles']
    cattrname = schema.get('csvAttr','')       # schemaからcsv側の項目名を求めます
    cvalue=''                                  # csvの値の領域を初期化します
    if cattrname=='':                          # csvAttrの指定が無かった
        return ''
    if cattrname not in head_row:              # csvの項目名に一致する項目はあるか確認します
        handles['error'].set_errorobj('02', f'## schema ## column name specified in schema not found. #attr#, column name: {cattrname}')
    cvalue = data_row[head_row.index(cattrname)]  # その項目の項目値をcvalueに取り出す
    scm_type = schema.get('type','')
    if scm_type == '':
        # print(f'## ret_csv_value. schema: ',schema)
        handles['error'].set_errorobj('03', f'## schema ## type parameter missing in schema. #attr#, column name: {cattrname}')
    if schema['type'] == 'string':
        cvalue = str(cvalue)
        if schema.get('pattern') != None:
            cvalue = pattern_value(run, cvalue, schema['pattern'])
        if schema.get('format') != None:
            cvalue = format_value(run, cvalue, schema['format'])
    elif schema['type'] == 'number':
        if CT_text.isnumeric(cvalue):
            cvalue = CT_text.float(cvalue)
        else:
            handles['trace_log'].set_log(0,'04',f'invalid numeric value. #attr#, value: {str(cvalue)}')
    elif schema['type'] == 'integer':
        if cvalue.isdigit():
            cvalue = int(cvalue)
        else:
            handles['trace_log'].set_log(0,'05',f'invalid integer value. #attr#, value: {str(cvalue)}')
    elif schema['type'] == 'boolean':
        if cvalue in run['param']['truevalue']:
            cvalue = True
        elif cvalue in run['param']['falsevalue']:
            cvalue = False
        else:
            handles['trace_log'].set_log(0,'02','unknown boolean value: {str(cvalue)}')
    return cvalue                              # csvから求めた値を返却します

def format_value(run, value, format):
    global mapdic , prevattrs, leaf_obj, struc
    handles = run['handles']
    if (format == 'date'):
        date=value.split('/')
        if (len(date) == 1):
            return(date[0])
        elif (len(date) == 3):
            if (len(date[0]) == 2):
                date[0] = '20'+date[0]
            elif (len(date[0]) != 4):
                handles['trace_log'].set_log(0,'01','unknown date format: {value}')
            date[1] = ('0'+date[1])[-2:]
            date[2] = ('0'+date[2])[-2:]
            return(date[0]+'-'+date[1]+'-'+date[2])
        else:
            handles['trace_log'].set_log(0,'02','unknown date format: {value}')
    else:
        return(value)
def pattern_value(run, value, pattern):
    handles = run['handles']
    if (pattern[:7] == '^[0-9]{' and pattern[-2:] == '}$'):
        digits = int(pattern[7:-2])
        value = ('0'*digits + value)[-digits:]
    else:
        handles['trace_log'].set_log(0,'03','unknown patter specification: pattern: {pattern}')
    return(value)

def format_object(run, value, format):
    if (format == 'PostalAddress'):
        text = value['streetAddress']
        address.parse(run,text, value)
    else:
        return(value)

## ここからメインプログラムです
if __name__ == '__main__':
    ## まず、入出力ファイルの格納位置とファイル名を指定します。環境に応じて書き換えるか、何等かの方法
    ## で外部から渡してもらう様に変更する必要があります
    print (f'##### start ctoj converter #####')
    ## 変数を初期化します
    run = initialize_run()
    param = run['param']
    handles =run['handles']
    handles['error'] = err = error_log(run)
    handles['trace_log'] = trace_log(run)
    handles['entity'] = entity(run)
    
    ## 必須パラメータをチェックします
    if (param['inputfile'] == ''):
        err.set_errorobj('01', f'## error ## inputfile parameter missing')
    if (param['outputfile'] == ''):
        err.set_errorobj('02', f'## error ## outputfile parameter missing')
    if (param['schemafile'] == ''):
        err.set_errorobj('03', f'## error ## schemafile parameter missing')
    
    ## 拡張JSONスキーマを辞書として読み込みます
    handles['schema'] = exschema(run)
    # print(json.dumps(run['scm'],ensure_ascii=False, indent=2))

    ## inputfileのcsv文を二次元のリストとして読み込みます
    directory = run['param']['inputdir']
    filename = run['param']['inputfile']
    try:
        inputtable = CT_csv.load(directory, filename)
    except FileNotFoundError as e:
        err.set_errorobj('04', f'## error ## inputfileが見つからない. filename: {filename}, directory: {directory}')
    except UnicodeDecodeError as e:
        err.set_errorobj('05', f'## error ## inputfileのコードがUTF-8ではない. filename: {filename}, directory: {directory}')
    except csv.Error as e:
        err.set_errorobj('06', f'## error ## inputfileのcsvにエラーが見つかりました. filename: {filename}, directory: {directory}')
    except Exception as e:
        err.set_errorobj('07', f'## error ## inputfileのcsv読み込み時にエラーが発生しました. filename: {filename}, directory: {directory}, {str(e)}')
    handles['trace_log'].set_log(0,'08','inputdir: '+directory+', inputfile: '+filename)
    
    ## 各Entityのファイルを辞書のリストとして読み込みます
    try:
        for entityname in run['entitynames']:
            directory = entityname+'dir'
            filename = entityname+'file'
            if ((filename) in param and param[filename] != ''):
                handles[entityname] = GET_entities(run,entityname)
                handles[entityname].load(param[entityname + 'dir'],param[entityname + 'file'])
                handles['trace_log'].set_log(0,'09',f'referenced entities. directory: {directory}, {entityname}file: {filename}')
    except FileNotFoundError as e:
        handles['error'].set_errorobj('10', f'## error ## Entities load  error : {e}\n')
    handles['trace_log'].set_log(0,'09','outputdir: '+param['outputdir']+', outputfile: '+param['outputfile'])
    
    # 日本の自治体名の一覧を二次元のlistに展開します
    directory = run['param']['addressdir']
    filename = run['param']['addressfile']
    try:
        address_list = CT_csv.load(directory, filename)
    except FileNotFoundError as e:
        err.set_errorobj('11', f'## error ## {filename} ファイルが見つからない. directory: {directory}')
    except UnicodeDecodeError as e:
        err.set_errorobj('12', f'## error ## {filename} csvファイルのコードがUTF-8ではない. directory: {directory}')
    except csv.Error as e:
        err.set_errorobj('13', f'## error ## csvファイルにエラーが見つかりました. filename: {filename}, directory: {directory}')
    except Exception as e:
        err.set_errorobj('14', f'## error ## csvファイルの読み込み時にエラーが発生しました. filename: {filename}, directory: {directory}.')
    
    address_list = address_list[1:]
    run['address_table'] = address.convert2dict(address_list)
    run['address_index'] = address.convert2index(address_list)

    ## 拡張json schemaに記述されている順番に従ってinputfileの二次元listからからEntityのlistに変換する
    outlist=[]
    head_row =inputtable[0]
    for dataline  in range(1,len(inputtable)): # 入力csvの各行に対し処理します
        # ひとつのEntity分の辞書をリストの要素として追加します
        if inputtable[dataline] == []:
            continue
        handles['trace_log'].set_log(0,'10',f'CSV Line# {dataline+1}')
        entity = put_entity(run,head_row , inputtable[dataline])
        if not run['scm']['need_merge']:       # mergeの指定がない
            outlist.append(entity)             # 単純に末尾に追加します
        else:
            run['handles']['entity'].merge(outlist, entity, run['scm'])
    
    ## 最後に、jsontypeに合わせてEntityのリストをJSON文に変換して整形します
    if (param['jsontype'] == 'batch'):
        outdic={'actionType': 'append'}
        outdic['entities']=outlist
        outtext = json.dumps(outdic, indent=2, ensure_ascii=False)
    elif (param['jsontype']=='array'):
        outdic={}                              # JSONに変換するため、一旦辞書に変換する
        outdic['a']=outlist
        outtext = json.dumps(outdic, indent=2, ensure_ascii=False)
        outtext = '  '+outtext[9:len(outtext)-2]+'\n'  # 大括弧の外側の文字を削除する
    elif (param['jsontype']=='entity'):
        outdic={}                              # JSONに変換するため、一旦辞書に変換する
        outdic['a']=outlist
        outtext = json.dumps(outdic, indent=2, ensure_ascii=False)
        outtext = outtext[11:len(outtext)-5]   # 大括弧の外側の文字を削除する
    else:
        handles['error'].set_errorobj( '02', f'invalid jsontype option\n')
    
    ## 出力します
    directory = run['param']['outputdir']
    filename = run['param']['outputfile']
    try:
        CT_text.save(directory, filename, outtext)
    except FileNotFoundError as e:
        err.set_errorobj('15', f'#F# error ## {filename} ファイルが見つからない. directory: {directory}')
    except PermissionError as e:
        err.set_errorobj('16', f'## error ## outputfileが他のプログラムから使われている. または利用権がない. filename: {filename}, directory: {directory}')
    
    if (run['errorflag']==True):
        handles['trace_log'].set_log(0,'11',f'CSV to JSON translation faileded')
        print('Error occurd. Please, check message, parameters and files\n')
        input(run['errorobj'])
    if run['warning'] == True:
        handles['trace_log'].__del__()
        input(f'CSVからJSONへの変換中にワーニングが発生しました。ログを確認してください\n')
    else:
        handles['trace_log'].set_log(0,'12',f'CSV translated to JSON successfully. Enity number: '+str(len(outlist)))
        handles['trace_log'].__del__()
        print(f'CSV translated to JSON successfully. output Entity number: {str(len(outlist))}\n')