#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# checked with 3.10,3.11
# Created by H. Inoue on 2024-02-10 
# Updated by H. Inoue on 2024-02-26 
"""
NGSI v2に準拠するJson文に関係するクラス群です。
本クラス群はCoppellTechnologiesのツールから利用される事を前提としています。
runテーブルを経由してツールやクラス間で情報共有します
"""
import os
import sys
import json
import inspect
from my_package.CoppellTechnologies_common import CT_json, CT_text
# import pdb

class GET_entities:
    ## Orionが起動していない環境で、Entityの検索機能を提供するクラスです
    
    ## コンストラクタでは、各種変数を受け取ります
    def __init__(self, run,entity_type):
        """
        コンストラクタはEntities (Entityを要素とするlist)毎に実行します
        entity_type: Entity typeを指定します
        """
        self.run = run                         # runテーブルをselfに保持します
        self.entity_type = entity_type
        self.entities = []                     # 検索対象のentityのリストを格納する領域です
        return
    
    ## 入力のJSON文を辞書に変換します。返却値は無く、インスタンスのentitiesに辞書を格納します
    def load (self, directory, filename)->dict:
        handles = self.run['handles']
        err = handles['error']
        try:
            entities = CT_json.load(directory, filename)
        except FileNotFoundError as e:
            err.set_errorobj('01', f'## error ## {filename} ファイルが見つからない. directory: {directory}')
        except UnicodeDecodeError as e:
            err.set_errorobj('02', f'## error ## {filename} jsonファイルのコードがUTF-8ではない. directory: {directory}')
        except json.Error as e:
            err.set_errorobj('03', f'## error ## jsonファイルにエラーが見つかりました. filename: {filename}, directory: {directory}')
        except Exception as e:
            err.set_errorobj('04', f'## error ## jsonファイルの読み込み時にエラーが発生しました. filename: {filename}, directory: {directory}.')
        # entitiesの形式を確認して、リストの先頭を探します
        if isinstance(entities,list):          # list形式にentityが並んでいる形式
            pass
        elif isinstance(entities,dict):        # entity単独かバッチ登録可能な形式
            if entities.get('entities'):       # バッチ形式
                entities = entities['entities']
            else:                              # 単独entity
                entities = [entities]
        else:
            err.set_errorobj('11', f'## error ## jsonファイルの形式が正しくない. filename: {filename}, directory: {directory}.')
        if not len(entities):
            err.set_errorobj('12', f'## error ## jsonファイルから読み込んだentityがゼロ件. filename: {filename}, directory: {directory}.')
        if not entities[0].get('id') or not entities[0].get('type'):
            err.set_errorobj('13', f'## error ## jsonファイルから読み込んだentityにidまたはtypeが無い. filename: {filename}, directory: {directory}.')
        if entities[0]['type'] != self.entity_type:
            err.set_errorobj('14', f'## error ## jsonファイルから読み込んだentity typeが指定したtypeと異なる. filename: {filename}, directory: {directory}.')
        self.entities = entities

    ## 指定されたEntityのAttribute name, Attributte valueからEntityを求め、辞書形式で返却します
    def get (self,
             key: str,
             value)->dict:
        """
        Entityのlistを順番に検索し、Attributeのnameとvalueが一致するEntityを返却します
        key: Attribute name
        value: Attributeのvalue
            strまたはintの時には、単純に値が一致するAttributeを持つEntityを探します
            dictの時には、
        """
        if isinstance(value,str) or isinstance(value,int):
            for outer_dict in self.entities:
                if key in outer_dict:
                    if isinstance(outer_dict[key], dict):
                        if outer_dict[key]['value'] == value:
                            return(outer_dict)
                    else:
                        if outer_dict[key] == value:
                            return(outer_dict)
            return {}
        elif isinstance(value,dict):
            for outer_dict in self.entities:
                if key in outer_dict:
                    if isinstance(outer_dict[key], dict):
                        inner_dict=outer_dict[key]['value']
                        for subkey, subvalue in value.items():
                            if subkey in inner_dict and inner_dict[subkey] == subvalue:
                                continue
                        return(outer_dict)
        return({})
    
    def get_subattr (self, key1, key2, value2):
        # print(f'## get_subattr: key1: {key1}, key2: {key2}, value2: {value2}')
        # print(f'                entities: ', self.entities)
        for outer_dict in self.entities:
            if (key1 in outer_dict and
                key2 in outer_dict[key1]['value'] and
                outer_dict[key1]['value'][key2] == value2):
                return(outer_dict)
        return {}
    
    def get_listed_subattr (self, key1, key2, value2, key3, value3):
        for outer_dict in self.entities:
            if key1 in outer_dict and isinstance(outer_dict[key1]['value'], list):
                for inner_dict in outer_dict[key1]['value']:
                    if key2 in inner_dict and inner_dict[key2] == value2:
                        if key3 in inner_dict and inner_dict[key3] == value3:
                            return(outer_dict)
        return {}

class POST_entities:
    ## Orionが起動していない環境で、Entityの検索機能を提供するクラスです
    
    ## コンストラクタでは、各種変数を受け取ります
    def __init__(self, entity_type, options):
        self.options = options                 # optionsのアドレスをselfに保持します
        return
    
    def save(self, directory, filename, text):
    ## テキストをファイルとして出力します
        if directory != '':
            try:
                os.chdir(directory)
            except FileNotFoundError as e:
                self.options['handles']['error'].set_errorobj(self.options,'01',
                    f'## error ## {directory} directory does not exist.')
                raise e
            except Exception as e:
                self.options['handles']['error'].set_errorobj(self.options,'02', 
                    f'## error ## change directory error. directory: {directory}.')
                raise e
        
        try:
            f=open(filename, 'w', encoding='utf-8')
        except PermissionError as e:
            self.options['handles']['error'].set_errorobj(self.options,'03', 
                f'## error ## {filename} file is already used or not permitted. dir : {directory}')
            raise e
        except Exception as e:
            self.options['handles']['error'].set_errorobj(self.options,'03',
                f'## error ## open error. filename: {filename}, code: {e}')
            raise e
        
        n = f.write(text)                          # nは使いません。nは書き込んだ文字数です
        f.close()
        return

class exschema:
    """
    export用の拡張Json schemaに関する処理です。
    """
    def __init__(self,
                 run: dict)->None:
        """
        指定された拡張Json Schemsをテキストファイルから読込み、dictとしてメモリに展開します。
        展開したdictはrun['scm']に格納します
        run: ツールのrunテーブル
        """
        self.run = run
        err = run['handles']['error']
        param = run['param']
        directory = param['schemadir']
        filename = param['schemafile']
        
        ## json schemaを読み出します
        try:
            schema_dict = CT_json.load(directory,filename)
        except FileNotFoundError as e:
            err.set_errorobj('01', f'## error ## {filename} does not exist. dir: {directory}')
        except json.JSONDecodeError as e:
            err.set_errorobj('02', f'## error ## json decoding error in {filename} : {e}')
        except Exception as e:
            err.set_errorobj('03', f'## error ## schemaの読み込みでエラーが発生しました. dir: {directory}, file: {filename}, error: {str(e)}')
        
        # 処理不要なattributeを削除します。
        if schema_dict.get('type') == None or schema_dict['type'] != 'object':
            err.set_errorobj('01', f'## error ## {filename} type MUST object in schema')
        self.delete_properties(schema_dict)

        # 最後に実行用のスキーマの内容をチェックします
        self.check_schema(schema_dict)
        run['scm'] = schema_dict
        return
    
    def delete_properties(self,
                          schema: dict):
        """
        schemaは拡張スキーマのトップのdict
        """
        def delete_object(schema: dict,
                          flags: dict):
            """
            schemaは拡張スキーマのtypeがobjectであるdict
            """
            prop = schema.get('properties',[])
            prop_copy = prop.copy()
            for key,value in prop_copy.items():    # sub-propertiesのループ
                value_type = value.get('type','')
                if value_type == 'object':
                    flags_sub = {'ctoj':False, 'const':False}              # 参照渡しにすためdict
                    delete_object(value,flags_sub)
                    msg =  f'flags_sub: ' + str(flags_sub['ctoj']) + f', key {key}, value: ' + str(value)
                    if flags_sub['ctoj'] == True:
                        flags['ctoj'] = True
                    else:
                        if (not flags_sub['const'] and
                            schema.get('required') != None and key not in schema['required']):
                            del prop[key]
                elif value_type == 'array':
                    delete_array(value,flags)
                else:
                    flags_sub = {'ctoj':False, 'const':False}              # 参照渡しにすためdict
                    for k1 in value.keys():
                        if k1 in ('dataSource', 'dataTarget'):
                            flags_sub['ctoj'] = True
                        if k1 == 'const':
                            flags_sub['const'] = True
                    if flags_sub['ctoj']:
                        flags['ctoj'] = True
                    elif not flags_sub['const']:
                        del prop[key]
            return
        
        def delete_array(schema: dict,
                         flags: dict):
            """
            schemaは拡張スキーマのtypeがarrayであるdict
            """
            items = schema.get('items',[])
            if items == []:
                err.set_errorobj('01', f'## schema ## typeがarrayなのにitemsの記述がない')
            if isinstance(items,list):             # バリューが複数変換される
                for i1 in items:                   # items内のリスト毎のループ
                    items_type = i1.get('type','')
                    if items_type == '':
                        err.set_errorobj('02', f'## schema ## typeの記述がない')
                    elif items_type == 'object':
                        delete_object(i1,flags)
                    elif items_type == 'array':
                        delete_array(i1,flags)
                    else:
                        for k1 in i1.keys():
                            if k1 in ('dataSource', 'dataTarget'):
                                flags['ctoj'] = True
                                return
            else:                                  # 順序無しリスト
                items_type = items.get('type','')
                if items_type == 'object':
                    delete_object(items,flags)
                elif items_type == 'array':
                    delete_array(items,flags)
                else:
                    for k1 in items.keys():
                        if k1 in ('dataSource', 'dataTarget'):
                            flags['ctoj'] = True
                            return
            return
        
        run = self.run
        err = run['handles']['error']
        required = schema.get('required',[])
        if not isinstance(required,list):
            msg = f'## error ## required MUST list. required: ' + str(required)
            err.set_errorobj('02', msg)
        prop = schema.get('properties',{})
        prop_copy = prop.copy()
        for key,value in prop_copy.items():    # propertiesのループ
            run['errorinfo']['attr'] = key
            if key in required:
                continue
            flags = {'ctoj':False}             # 参照渡しにすためdict
            value_type = value.get('type','')
            if value_type == 'object':
                delete_object(value,flags)
            elif value_type == 'array':        # NGSI V2のnormalized形式の仕様上あり得ない
                err.set_errorobj('01', f'## schema ## attributeの形式にarrayが現れた')
                delete_array(value,flags)
            else:                              # NGSI V2のnormalized形式の仕様上あり得ない
                err.set_errorobj('02', f'## schema ## idとtype以外のattributeの形式に即値が現れた')
                for k1 in value.keys():
                    if k1 in ('dataSource', 'dataTarget'):
                        flags['ctoj'] = True
                        break
            if flags['ctoj'] == False:
                del prop[key]
        return
    
    def check_schema(self,schema:dict)->None:
        """
        拡張Json schemaの記述をチェックすると共に、形式を整えて後段の処理を
        効率化します。
        schema: 拡張jsonスキーマのトップのdict
        """
        def check_object(schema: dict,
                         flags: dict)->None:
            """
            schema: typeがobjectであるdict
            """
            if schema.get('mergekey'):
                err.set_errorobj('01', f'## schema ## mergekeyの記述位置はtypeがarrayのオブジェクトの中です')
            if schema.get('oneOf'):
                err.set_errorobj('02', f'## schema ## oneOfの指定はサポートしていません')
            if any(item in schema for item in ('csvAttr','entityType','searchAttr','valueAttr','identificationType','prefix')):
                err.set_errorobj('03', f'## schema ## csvAttr,entityType,searchAttr,valueAttr,identificationType,prefixはdataSource内に記述します')
            prop = schema.get('properties',{})
            if prop == {}:
                err.set_errorobj('04', f'## schema ## typeがobjectなのにpropertiesの指定がない')
            if schema.get('required') != None:
                for attr in schema['required']:
                    if not prop.get(attr):
                        err.set_errorobj('05', f'## schema ## requiredとAttribute名の対応が取れない。required: {attr}')
            
            for prop_key,attr_value in prop.items():
                # type指定のチェック
                attr_type = attr_value.get('type','')
                if attr_type == '':
                    if not attr_value.get('const'):
                        msg = f'Key: {prop_key}, Value: {attr_value}'
                        err.set_errorobj('07', f'## schema ## type指定が無い. {msg}')
                    const = attr_value['const']
                    if isinstance(const,str):
                        attr_type = 'string'
                    elif isinstance(const,bool):
                        attr_type = 'boolean'
                    elif isinstance(const,(int,float)):
                        attr_type = 'number'
                    else:
                        err.set_errorobj('08', f'## schema ## constに指定した値の型が正しくない. Attribute: {prop_key}')
                    attr_value['type'] = attr_type
                # ネスト先のチェック
                if attr_type == 'object':
                    check_object(attr_value,flags)
                elif attr_type == 'array':
                    check_array(attr_value,flags)
                else:
                    # 各AttributeのdataSource指定の確認
                    check_dataSource(attr_value)
                    if attr_type not in ('integer','number','string','boolean'):
                        err.set_errorobj('09', f'## schema ## typeに指定できる型はobject,array,string,boolean,number,integerです. #attr#, type: {str(attr_type)}')
            return
        
        def check_array(schema: dict,
                        flags: dict)->None:
            """
            schema: typeがarrayであるdictです
            """
            mergekey = schema.get('mergekey','')
            ordered = schema.get('ordered',False)
            # まずは、全体の形式チェック
            if schema.get('oneOf'):
                err.set_errorobj('07', f'## schema ## oneOfの指定はサポートしていません')
            if any(item in schema for item in ('csvAttr','entityType','searchAttr','valueAttr','identificationType','prefix')):
                err.set_errorobj('08', f'## schema ## csvAttr,entityType,searchAttr,valueAttr,identificationType,prefixはdataSource内に記述します')
            items = schema.get('items','')
            if not items:
                err.set_errorobj('09', f'## schema ## typeがarray指定なのにitemsの指定がない')
            # mergekeyのチェック
            if not flags['merge'] and mergekey != '':  # merge指定のないAttributeにmergekeyの指定があった
                err.set_errorobj('20', f'## schema ## merge指定のないAttributeにmergekeyの指定がある')
            if mergekey == '':                 # mergekeyの指定が無いリスト
                if flags['merge'] and not ordered: # concatinate mergeが見つかった
                    flags['concatinate'] = True    # concatinate mergeのフラグをTrue
            else:                              # mergekeyの指定がある
                if flags['concatinate']:       # 順序無しリストの先にマージの指定があった
                    err.set_errorobj('21', f'## schema ## mergekeyの指定の上位に指定のないArrayがある.')
                if isinstance(mergekey,str):   # Attoribute名を指定された
                    if items['type'] != 'object':
                        err.set_errorobj('22', f'## schema ## mergekeyの項目名指定はarrayの要素がobjectの場合だけです')
                    prop = items.get('properties','')
                    if prop == '':
                        err.set_errorobj('23', f'## schema ## mergekey指定の下にpropertiesの指定がない')
                    if prop.get(mergekey) == None:
                        err.set_errorobj('24', f'## schema ## mergekeyに指定した項目名が直下のsub-attributeに見つかりません')
                elif isinstance(mergekey,list):
                    if items['type'] != 'object':
                        err.set_errorobj('25', f'## schema ## mergekeyの項目名指定arrayの要素がobjectの場合だけです')
                    if len(mergekey) != 2:
                        err.set_errorobj('26', f'## schema ## mergekeyのリストの長さは、2だけです')
                    mk0 = mergekey[0]
                    if not isinstance(mk0,str):
                        err.set_errorobj('27', f'## schema ## mergekeyのリストの一つ目は文字列です')
                    mk0_prop = items.get('properties','')
                    mk0_attr_value = mk0_prop.get(mk0,'')    # mergekeyに指定されたattributeのvalue
                    if mk0_attr_value == '':
                        err.set_errorobj('28', f'## schema ## mergekeyに指定したsubattribute名が見つかません. attribute name: {mk0}')
                    mk0_prop_type = mk0_attr_value.get('type','')
                    if mk0_prop_type == '' or mk0_prop_type != 'array':
                        err.set_errorobj('29', f'## schema ## mergekeyに指定したsubattributeがarrayではない. attribute name: {mk0}')
                    mk1 = mergekey[1]
                    if not isinstance(mk1,int):
                        err.set_errorobj('30', f'## schema ## mergekeyのリストの二つ目は整数です')
                    mk1_items = mk0_attr_value['items']
                    if not isinstance(mk1_items,list):
                        err.set_errorobj('31', f'## schema ## mergekeyのリスト指定に対し、指定したsubattributeのitemsがリスト構造ではない')
                else:                              # マージキーの指定が未知
                    err.set_errorobj('32', f'## schema ## mergekeyの指定方法が正しくありません')
            # itemsの形式チェック
            if isinstance(items,list):
                for i0 in items:
                    check_array_sub(i0, flags)
            elif isinstance(items,object):
                check_array_sub(items,flags)
                """
                items_type = items.get('type','')
                if items_type == '':
                    err.set_errorobj('12', f'## schema ## items内にtype指定がない')
                if items_type == 'object':
                    check_object(items,flags)
                elif items_type == 'array':
                    check_array(items,flags)
                else:
                    if items_type not in ('number','integer','string','boolean'):
                        err.set_errorobj('13', f'## schema ## items内のtype指定が正しくない')
                    check_dataSource(items)
                """
            else:
                err.set_errorobj('14', f'## schema ## itemsの形式が正しくない')
            return
        
        def check_array_sub(items: dict, flags)->None:
            """
            schemaは拡張スキーマのtypeがarrayであるのitemsの要素のdict
            """
            items_type = items.get('type','')
            if not items_type:
                err.set_errorobj('10', f'## schema ## items内にtype指定がない')
            if items_type == 'object':
                check_object(items,flags)
            elif items_type == 'array':
                check_array(items,flags)
            else:
                if items_type not in ('number','integer','string','boolean'):
                    err.set_errorobj('11', f'## schema ## items内のtype指定が正しくない')
                check_dataSource(items)
            """
            items_type = items.get('type','')
            if not items_type:                 # typeを省略して良いのは値指定の時だけ
                err.set_errorobj('01', f'## schema ## items記述内にtype記述がない')
            elif items_type == 'object':
                check_object(items)
            elif items_type == 'array':
                if not flags['ordered']:
                    err.set_errorobj('02', f'## schema ## 多次元のarrayはorderedのみサポートしています')
                check_array(items)
            else:                      # 値の型指定が、指定がないか
                pass
            if flags['ordered']:
                if not items.get('order',''):
                    err.set_errorobj('02', f'## schema ## ordered指定のitems記述内にorder記述がない')
                if not isinstance(items['order'], int) or items['order'] < 0:
                    err.set_errorobj('03', f'## schema ## orderの指定は整数でなくてはならない')
            else:                      # 検索用のキーの順番を整理しておく
                prop = items['properties']
                found = ''
                for k1,v1 in prop.items(): # propertiesの要素をサーチ
                    if not isinstance(v1,dict):
                        continue
                    if v1.get('dataTarget'):
                        if isinstance(v1['dataTarget'], list):
                            err.set_errorobj('04', f'## schema ## dataTargetのリスト定義はサポートしていません')
                        found = k1
                        break
                if not found:
                    return
                search_key = {}        # 検索用の情報を作り変えます
                for k1,v1 in prop.items():        # itemsの要素の中身をサーチ
                    if not isinstance(v1,dict):
                        continue
                    if v1.get('const'):
                        search_key[k1] = v1['const']
                        break
                if len(search_key) == 0:
                    err.set_errorobj('04', f'## schema ## 順序無しリストを検索する指定がない')
                items['searchKey'] = search_key 
            """
            return
        
        def check_dataSource(schema):
            # print(f'check_dataSource. schema: {str(schema)}')
            data_source = schema.get('dataSource','')
            if not data_source:
                a ={}
                if schema.get('const'):
                    schema['dataSource'] = {'type': schema['type'], 'const': schema['const']}
                if schema.get('entityType'):
                    err.set_errorobj('10', f'## schema ## entityTypeに対応するdataSourceの指定がない. #attr#')
                # dataSource = schema.get('dataSource','')
                # print(f'check_dataSource(no dataSource). schema: {str(dataSource)}')
                return
            if isinstance(data_source,list):
                for data_source_sub in data_source:
                    if not data_source_sub.get('type'):
                        data_source_sub['type'] = attr_value['type']
                    check_dataSource_sub(data_source_sub)
            elif isinstance(data_source,dict):
                if not data_source.get('type'):
                    data_source['type'] = schema['type']
                if schema.get('const') and not data_source.get('const'):
                    data_source['const'] = schema['const']
                if schema.get('format') and not data_source.get('format'):
                    data_source['format'] = schema['format']
                if schema.get('pattern') and not data_source.get('pattern'):
                    data_source['pattern'] = schema['pattern']
                check_dataSource_sub(data_source)
            else:
                # print(f'check_dataSource(error): schema: ',attr_value)
                err.set_errorobj('09', f'## schema ## dataSourceの形式が不正です')
        
        def check_dataSource_sub (schema):
            """
            dataSourceの内容をチェックします
            schema: 拡張スキーマのdataSourceやalternateのdict
            """
            if schema.get('dataSource'):
                data_source = schema['dataSource']
                if isinstance(data_source,list):
                    for data_source_sub in data_source:
                        if not data_source_sub.get('type'):
                            data_source_sub['type'] = schema['type']
                        check_dataSource_sub(data_source_sub)
                elif isinstance(data_source,dict):
                    if not data_source.get('type'):
                        data_source['type'] = schema['type']
                    check_dataSource_sub(data_source)
                else:
                    # print(f'check_dataSource_sub(error): schema: ',schema)
                    err.set_errorobj('01', f'## schema ## dataSourceの形式が不正です')
            elif schema.get('entityType'):
                err.set_errorobj('01', f'## schema ## entityTypeに対応するdataSourceの指定がない. #attr#')
            if schema.get('alternate'):
                alternate = schema['alternate']
                if isinstance(alternate,list):
                    for alternate_sub in alternate:
                        if not alternate_sub.get('type'):
                            alternate_sub['type'] = schema['type']
                        check_dataSource_sub(alternate_sub)
                elif isinstance(alternate,dict):
                    if not alternate.get('type'):
                        alternate['type'] = schema['type']
                    check_dataSource_sub(alternate)
                else:
                    # print(f'check_dataSource_sub(error): schema: ',alternate)
                    err.set_errorobj('02', f'## schema ## alternateの形式が不正です')
            return
        
        # check_schemaのメインです
        run = self.run
        err = run['handles']['error']
        flags = {'need_merge': False}
        # まずは最上位の形式チェック
        if schema.get('type') == None or schema['type'] != 'object':
            scm = str(schema)[:200]
            err.set_errorobj('01', f'## schema ## schemaの最上位にtypeがobjectである指定が無い. schema: {scm}')
        prop = schema.get('properties',{})
        if prop == {}:
            err.set_errorobj('02', f'## schema ## schemaの最上位にpropertiesが無い')
        if schema.get('merge') != None:
            err.set_errorobj('03', f'## schema ## schema内のmergeの指定箇所が正しくない')
        a2 = prop.get('id','') 
        if a2 == '':
            err.set_errorobj('04', f'## schema ## schemaに必須Attributeであるidの指定が無い')
        if a2.get('mergekey') != None:
            del a2['mergekey']
        a2 = prop.get('type','') 
        if a2 == '':
            err.set_errorobj('05', f'## schema ## schemaに必須Attributeであるtypeの指定が無い')
        if a2.get('mergekey') != None:
            del a2['mergekey']
        
        # requiredのチェック
        flags['concatinate'] = False           # セット(順序無しリスト)のフラグ初期化
        # requiedのattributeの存在確認
        if schema.get('required') != None:
            for attr in schema['required']:
                if not prop.get(attr):
                    err.set_errorobj('06', f'## schema ## requiredとAttribute名の対応が取れない。required: {attr}')
        # 各attributeの各種指定の確認
        for attr_name, attr_value in prop.items():
            run['errorinfo']['attr'] = attr_name
            # 各attributeのtype指定の確認
            attr_type = attr_value.get('type','')
            if attr_type == '':
                if attr_value.get('const') == None:
                    err.set_errorobj('07', f'## schema ## type指定が無い. Attribute: {attr_name}')
                const = attr_value['const']    # constの指定がある場合はtypeを省略できる
                if isinstance(const,str):
                    attr_type = 'string'
                elif isinstance(const,bool):
                    attr_type = 'boolean'
                elif isinstance(const,(int,float)):
                    attr_type = 'number'
                else:
                    err.set_errorobj('08', f'## schema ## constに指定した値の型が正しくない. #attr#')
                attr_value['type'] = attr_type
            if not isinstance(attr_type,str) or attr_type not in ('object', 'array', 'string', 'boolean', 'number','integer'):
                err.set_errorobj('09', f'## schema ## typeに指定できる型はobject,array,string,boolean,number,integerです. Attribute: {attr_name}, type: {str(attr_type)}')
            # 各Attributeのmerge指定の確認
            if attr_value.get('merge') == None:
                flags['merge'] = False         # Attribute全体にmerge:unionの指定が無い
            else:
                if attr_value['merge'] != 'union':
                    err.set_errorobj('11', f'## schema ## mergeモードはunonのみサポートしています')
                flags['merge'] = True
                flags['need_merge'] = True     # Attribute全体にmerge:unionの指定がある
            # 各attribute内のネストの確認
            if attr_type == 'object':
                check_object(attr_value,flags)
            elif attr_type == 'array':
                check_array(attr_value,flags)
            else:                              # typeがnumber,string等
                # 各AttributeのdataSource指定の確認
                check_dataSource(attr_value)
        schema['need_merge'] = flags['need_merge']
        return
        
class imschema:
    def __init__(self,
                 run: dict)->dict:
        """
        指定されたimport用拡張Json Schemを読込み、メモリに展開します
        """
        self.run = run
        err = run['handles']['error']
        param = run['param']
        directory = param['schemadir']
        filename = param['schemafile']
                
        ## json schemaを読み出します
        
        try:
            schema_dict =  CT_json.load(directory,filename)
        except FileNotFoundError as e:
            err.set_errorobj('01', f'## error ## {filename} does not exist. dir: {directory}')
        except json.JSONDecodeError as e:
            err.set_errorobj('02', f'## error ## json decoding error in {filename} : {e}')
        except Exception as e:
            err.set_errorobj('03', f'## error ## schemaの読み込みでエラーが発生しました. dir: {directory}, file: {filename}, error: {str(e)}')
        
        # 不要なattributeを削除します。
        self.delete_properties(schema_dict)
        
        # 最後に実行用のスキーマの内容をチェックします
        self.check_schema(schema_dict)
        run['scm'] = schema_dict
        return
    
    def delete_properties(self,
                          schema: dict):
        """
        schemaは拡張スキーマのトップのdict
        """
        def delete_object(schema: dict,
                          flags: dict):
            """
            schemaは拡張スキーマのtypeがobjectであるdict
            """
            prop = schema.get('properties',{})
            if not prop:
                err.set_errorobj('03', f'## schema ## typeがobjectなのにpropertiesがない')
            prop_copy = prop.copy()
            for key,value in prop_copy.items():    # sub-propertiesのループ
                value_type = value.get('type','')
                if value_type == 'object':
                    flags_sub = {'jtoc':False, 'const':False}              # 参照渡しにすためdict
                    delete_object(value,flags_sub)
                    if flags_sub['jtoc']:
                        flags['jtoc'] = True
                    else:
                        if (not flags_sub['const']):
                            del prop[key]
                elif value_type == 'array':
                    delete_array(value,flags)
                else:
                    flags_sub = {'jtoc':False, 'const':False}              # 参照渡しにすためdict
                    for k1 in value.keys():
                        if k1 == 'dataTarget':
                            flags_sub['jtoc'] = True
                        if k1 == 'const':
                            flags_sub['const'] = True
                    if flags_sub['jtoc']:
                        flags['jtoc'] = True
                    elif not flags_sub['const']:
                        del prop[key]
            return
        
        def delete_array(schema: dict,
                          flags: dict):
            """
            schemaは拡張スキーマのtypeがarrayであるdict
            """
            items = schema.get('items',[])
            if items == []:
                err.set_errorobj('01', f'## schema ## typeがarrayなのにitemsの記述がない')
            if isinstance(items,list):         # バリューが複数変換される
                for i1 in items:               # items内のリスト毎のループ
                    items_type = i1.get('type','')
                    if not items_type:
                        if i1.get('properties'):
                            err.set_errorobj('01', f'## schema ## items記述内にtype記述がない')
                    elif items_type == 'object':
                        delete_object(i1,flags)
                    elif items_type == 'array':
                        delete_array(i1,flags)
                    else:                      # 値の型指定が、指定がないか、 
                        for k1 in i1.keys():
                            if k1 == 'dataTarget':
                                flags['jtoc'] = True
                                return
            else:                              # 順序無しリスト
                items_type = items.get('type','')
                if not items_type:
                    if items.get('properties'):
                        err.set_errorobj('01', f'## schema ## items記述内にtype記述がない')
                elif items_type == 'object':
                    delete_object(items,flags)
                elif items_type == 'array':
                    delete_array(items,flags)
                else:
                    for k1 in items.keys():
                        if k1 == 'dataTarget':
                            flags['jtoc'] = True
                            return
            return
    
        run = self.run
        err = run['handles']['error']
        prop = schema.get('properties',{})
        if prop == {}:
            print(f'delete_properties(error): schema: ',schema)
            err.set_errorobj('01', f'## error ## 拡張schemaの最上位にpropertiesが無い')
        prop_copy = prop.copy()
        for key,value in prop_copy.items():    # propertiesのループ
            run['errorinfo']['attr'] = key
            flags = {'jtoc':False}             # 参照渡しにすためdict
            value_type = value.get('type','')
            if value_type == 'object':
                delete_object(value,flags)
            elif value_type == 'array':        # NGSI V2のnormalized形式の仕様上あり得ない
                err.set_errorobj('02', f'## schema ## 拡張schemaがnormalized形式ではない')
            else:                              # NGSI V2のnormalized形式の仕様上あり得ない
                for k1 in value.keys():
                    if k1 == 'dataTarget':
                        flags['jtoc'] = True
                        break
            if flags['jtoc'] == False:
                del prop[key]
        return
    
    def check_schema(self,schema):
        """
        拡張Json schemaの記述をチェックすると共に、形式を整えて後段の処理を
        効率化します。
        schema: 拡張jsonスキーマのトップのdict
        """
        def check_object(schema: dict)->None:
            """
            schemaは拡張スキーマのtypeがobjectであるdict
            """
            if schema.get('oneOf'):
                err.set_errorobj('02', f'## schema ## oneOfの指定はサポートしていません')
            if any(item in schema for item in ('csvAttr','entityType','searchAttr','valueAttr','identificationType','prefix')):
                err.set_errorobj('03', f'## schema ## csvAttr,entityType,searchAttr,valueAttr,identificationType,prefixはdataSource内に記述します')
            prop = schema.get('properties',{})
            if not prop:
                err.set_errorobj('03', f'## schema ## typeがobjectなのにpropertiesがない')
            for prop_key,attr_value in prop.items():    # sub-propertiesのループ
                # print(f'## check_object(top): properties: key: {key}, value: ', value)
                # type指定のチェック
                attr_type = attr_value.get('type','')
                if attr_type == '':
                    if not attr_value.get('const'):
                        msg = f'Key: {prop_key}, Value: {attr_value}'
                        err.set_errorobj('07', f'## schema ## type指定が無い. {msg}')
                    const = attr_value['const']
                    if isinstance(const,str):
                        attr_type = 'string'
                    elif isinstance(const,bool):
                        attr_type = 'boolean'
                    elif isinstance(const,(int,float)):
                        attr_type = 'number'
                    else:
                        err.set_errorobj('08', f'## schema ## constに指定した値の型が正しくない. Attribute: {prop_key}')
                    attr_value['type'] = attr_type
                # ネスト先のチェック
                if attr_type == 'object':
                    check_object(attr_value)
                elif value_type == 'array':
                    check_array(attr_value)
                else:
                    # 各AttributeのdataSource指定の確認
                    check_dataTarget(attr_value)
                    if attr_type not in ('integer','number','string','boolean'):
                        err.set_errorobj('09', f'## schema ## typeに指定できる型はobject,array,string,boolean,number,integerです. #attr#, type: {str(attr_type)}')
                # print(f'## check_object(end): properties: key: {key}, value: ', value)
            return
        
        def check_array(schema: dict)->None:
            """
            schemaは拡張スキーマのtypeがarrayであるdict
            """
            if schema.get('oneOf'):
                err.set_errorobj('07', f'## schema ## oneOfの指定はサポートしていません')
            if any(item in schema for item in ('csvAttr','entityType','searchAttr','valueAttr','identificationType','prefix')):
                err.set_errorobj('08', f'## schema ## csvAttr,entityType,searchAttr,valueAttr,identificationType,prefixはdataSource内に記述します')
            flags = {'ordered' : schema.get('ordered',False)}
            items = schema.get('items',[])
            if not items:
                err.set_errorobj('09', f'## schema ## typeがarrayなのにitemsの記述がない')
            # itemsの形式チェック
            if isinstance(items,list):         # バリューが複数変換される
                for i0 in items:               # items内のリスト毎のループ
                    check_array_sub(i0, flags)
            elif isinstance(items,object):
                check_array_sub(items, flags)
            else:                              # itemsの指定が単独
                err.set_errorobj('14', f'## schema ## itemsの形式が正しくない')
            return
        
        def check_array_sub(items: dict, flags)->None:
            """
            schemaは拡張スキーマのtypeがarrayであるのitemsの要素のdict
            """
            items_type = items.get('type','')
            if not items_type:                 # typeを省略して良いのは値指定の時だけ
                err.set_errorobj('01', f'## schema ## items記述内にtype記述がない')
            elif items_type == 'object':
                check_object(items)
            elif items_type == 'array':
                if not flags['ordered']:
                    err.set_errorobj('02', f'## schema ## 多次元のarrayはorderedのみサポートしています')
                check_array(items)
            else:                      # 値の型指定が、指定がないか
                pass
            if flags['ordered']:
                if not items.get('order',''):
                    err.set_errorobj('02', f'## schema ## ordered指定のitems記述内にorder記述がない')
                if not isinstance(items['order'], int) or items['order'] < 0:
                    err.set_errorobj('03', f'## schema ## orderの指定は整数でなくてはならない')
            else:                      # 検索用のキーの順番を整理しておく
                prop = items['properties']
                found = ''
                for k1,v1 in prop.items(): # propertiesの要素をサーチ
                    if not isinstance(v1,dict):
                        continue
                    if v1.get('dataTarget'):
                        if isinstance(v1['dataTarget'], list):
                            err.set_errorobj('04', f'## schema ## dataTargetのリスト定義はサポートしていません')
                        found = k1
                        break
                if not found:
                    return
                search_key = {}        # 検索用の情報を作り変えます
                for k1,v1 in prop.items():        # itemsの要素の中身をサーチ
                    if not isinstance(v1,dict):
                        continue
                    if v1.get('const'):
                        search_key[k1] = v1['const']
                        break
                if len(search_key) == 0:
                    err.set_errorobj('04', f'## schema ## 順序無しリストを検索する指定がない')
                items['searchKey'] = search_key 
            return
        
        def check_dataTarget(schema):
            # print(f'check_dataTarget. schema: {str(schema)}')
            data_target = schema.get('dataTarget','')
            if not data_target:
                a ={}
                if schema.get('const'):
                    schema['dataTarget'] = {'type': schema['type'], 'const': schema['const']}
                if schema.get('entityType'):
                    err.set_errorobj('10', f'## schema ## entityTypeに対応するdataTargetの指定がない. #attr#')
                # dataTarget = schema.get('dataTarget','')
                # print(f'check_dataTarget(no dataTarget). schema: {str(dataTarget)}')
                return
            if isinstance(data_target,list):
                for data_target_sub in data_target:
                    if not data_target_sub.get('type'):
                        data_target_sub['type'] = data_target['type']
                    check_dataTarget_sub(data_target_sub)
            elif isinstance(data_target,dict):
                if not data_target.get('type'):
                    data_target['type'] = schema['type']
                if schema.get('const') and not data_target.get('const'):
                    data_target['const'] = schema['const']
                if schema.get('format') and not data_target.get('format'):
                    data_target['format'] = schema['format']
                if schema.get('pattern') and not data_target.get('pattern'):
                    data_target['pattern'] = schema['pattern']
                check_dataTarget_sub(data_target)
            else:
                # print(f'check_dataTarget(error): schema: ',attr_value)
                err.set_errorobj('09', f'## schema ## dataTargetの形式が不正です')
        
        def check_dataTarget_sub (schema):
            """
            dataTargetの内容をチェックします
            schema: 拡張スキーマのdataTargetのdict
            """
            if schema.get('dataTarget'):
                data_target = schema['dataTarget']
                if isinstance(data_target,list):
                    for data_target_sub in data_target:
                        if not data_target_sub.get('type'):
                            data_target_sub['type'] = schema['type']
                        check_dataTarget_sub(data_target_sub)
                elif isinstance(data_target,dict):
                    if not data_target.get('type'):
                        data_target['type'] = schema['type']
                    check_dataTarget_sub(data_target)
                else:
                    # print(f'check_dataTarget_sub(error): schema: ',schema)
                    err.set_errorobj('01', f'## schema ## dataTargetの形式が不正です')
            elif schema.get('entityType'):
                err.set_errorobj('01', f'## schema ## entityTypeに対応するdataTargetの指定がない. #attr#')
            if schema.get('alternate'):
                err.set_errorobj('02', f'## schema ## dataTargetでalternateはサポートしていません. #attr#')
            return
        
        run = self.run
        err = run['handles']['error']
        # まずは最上位の形式チェック
        prop = schema.get('properties',{})
        for key,value in prop.items():    # propertiesのループ
            run['errorinfo']['attr'] = key
            value_type = value.get('type','')
            if value_type == 'object':
                check_object(value)
            elif value_type == 'array':        # NGSI V2のnormalized形式の仕様上あり得ない
                err.set_errorobj('02', f'## schema ## 拡張schemaがnormalized形式ではない')
            else:                              # normalized形式の仕様上はidとtypeだけ
                pass
        
        pass                                   # 今後必要なチェックが見つかったら追加
        return
    
    def get_header(self,schema)->list:
        """
        schemaは、jsonスキーマのトップのdict
        返却値はcsvの項目名のリスト
        このメソッドはjtocからのみ呼ばれます
        """
        def get_header_object(schema: dict)->list:
            """
            schemaは拡張スキーマのtypeがobjectであるdict
            """
            ret_list = []
            prop = schema.get('properties',{})
            for key,value in prop.items():     # sub-propertiesのループ
                value_type = value.get('type','')
                if value_type == 'object':
                    ret_list += get_header_object(value)
                elif value_type == 'array':
                    ret_list += get_header_array(value)
                else:                          # typeが値
                    if value.get('dataTarget'):
                        ret_list += get_header_target(value['dataTarget'])
            return ret_list
        
        def get_header_array(schema: dict)->list:
            """
            schemaは拡張スキーマのtypeがarrayであるdict
            """
            items = schema.get('items',[])
            ret_list = []
            if items == []:
                err.set_errorobj('01', f'## schema ## typeがarrayなのにitemsの記述がない')
            if isinstance(items,list):             # バリューが複数変換される
                for i1 in items:                   # items内のリスト毎のループ
                    items_type = i1.get('type','')
                    if items_type == 'object':
                        ret_list += get_header_object(i1)
                    elif items_type == 'array':
                        ret_list += get_header_array(i1)
                    else:                          # typeが値の型
                        if i1.get('dataTarget'):
                            ret_list += get_header_target(i1['dataTarget'])
            else:                                  # 順序無しリスト
                items_type = items.get('type','')
                if items_type == 'object':
                    ret_list += get_header_object(items)
                elif items_type == 'array':
                    ret_list += get_header_array(items)
                else:
                    if items.get('dataTarget'):
                        ret_list += get_header_target(items['dataTarget'])
            return ret_list
        
        def get_header_target(schema: dict)->list:
            """
            schemaは拡張スキーマのdataSourceの値であるdict
            """
            if schema.get('csvAttr'):
                return [schema['csvAttr']]
            target = schema.get('dataTarget','')
            if not target:
                return []
            ret_list = []
            if isinstance(target,list):        # バリューが複数変換される
                for target_sub in target:      # items内のリスト毎のループ
                    ret_list += get_header_target(target_sub)
            else:                                  # 順序無しリスト
                ret_list += get_header_target(target)
            return ret_list
        
        run = self.run
        err = run['handles']['error']
        ret_list = []
        prop = schema.get('properties',{})
        for key,value in prop.items():    # propertiesのループ
            value_type = value.get('type','')
            if value_type == 'object':
                ret_list += get_header_object(value)
            elif value_type == 'array':        # keyValuesのサポート用にコードを残す
                ret_list += get_header_array(value)
            else:                              # keyValuesのサポート用にコードを残す
                for k1,v1 in value.items():
                    if k1 == 'dataTarget':
                        ret_list += [v1['csvAttr']]
        return ret_list
    
    def get_row(self,schema, inputjson)->list:
        """
        schemaは、jsonスキーマのトップのdict
        返却値はcsvの項目名のリスト
        """
        def get_object(schema: dict, inputjson)->list:
            """
            schema: 拡張スキーマのtypeがobjectであるdict
            inputjson: schemaに対応するdict
            """
            ret_list = []
            prop = schema.get('properties',{})
            for key,scm_value in prop.items():    # sub-propertiesのループ
                attrjson = {} if not inputjson else inputjson.get(key,{})
                value_type = scm_value.get('type','')
                if value_type == 'object':
                    ret_list += get_object(scm_value,attrjson)
                elif value_type == 'array':
                    ret_list += get_array(scm_value,attrjson)
                else:
                    if not scm_value.get('dataTarget'):
                        continue
                    dataTarget = scm_value['dataTarget']
                    ret_value = inputjson[key] if inputjson.get(key) else ''
                    ret_value = get_value(ret_value, dataTarget) if ret_value else ''
                    ret_list += [ret_value] if ret_value else ['']
            return ret_list
        
        def get_array(schema: dict,
                      inputjson)->list:
            """
            schemaは拡張スキーマのtypeがarrayであるdict
            """
            items_list = schema.get('items',{})
            ret_list = []
            if schema.get('ordered', False):   # 順序付きリストの指定がある
                if isinstance(items_list,list):    # リストの中から複数変換される
                    for items_dict in items_list:  # itemsのリストの要素毎のループ
                        if not items_dict['order']:
                            ret_list += ['']
                        index = items_dict.get('order')
                        if index >= len(inputjson):
                            ret_list += ['']
                        ret_list += get_array_order(items_dict,inputjson[index])
                else:                          # 単一のitems
                    items_dict = items_list
                    if not items_dict.get('order'):
                        ret_list += ['']
                    index = items_dict.get('order')
                    if index >= len(inputjson):
                        ret_list += ['']
                    ret_list += get_array_order(items_dict,inputjson[index])
            else:                              # 順序無しリスト
                # 順序無しリストは1次元のみサポート(check_schemaでチェック済み)
                if isinstance(items_list,list):    # リストの中から複数変換される
                    for items_dict in items_list:  # itemsのリストの要素毎のループ
                        ret_list += get_array_search(items_dict,inputjson)
                else:                                  # 順序無しリスト
                    ret_list += get_array_search(items_list,inputjson)
                return ret_list
        
        def get_array_order(schema: dict,
                      inputjson)->list:
            ret_list = []
            items_type = schema.get('type','')
            if items_type == 'object':
                ret_list += get_object(schema,inputjson)
            elif items_type == 'array':
                ret_list += get_array(schema,inputjson)
            else:                          # 値の型の指定 or 指定省略 
                if not schema.get('dataTarget'):
                    return ''
                ret_value = inputjson
                get_value(ret_value, schema['dataTarget'])
            return ret_list
        
        def get_array_search(schema: dict,inputjson)->list:
            """
            オブジェクトを要素とするArrayから、attribute名と値を与えて、値を求める
            schema: typeがarrayのitemsの要素。本メソッドが呼ばれる時、要素のtypeは必ずobject
            inputjson: inputfile内のobjectを要素とするarray
            """
            prop = schema['properties']
            # まずはdataTargetの指定があるAttributeを探します
            target = ''
            for k1,v1 in prop.items():
                if v1.get('dataTarget'):
                    target = k1
                    break
            if not target:
                return []
            search_key = schema['searchKey']   # constで指定した検索条件のリスト
            # inputjson中から条件に一致する最初の要素を取得
            result = next((item for item in inputjson if all(item[key] == value for key, value in search_key.items())),None)
            if not result:
                return ['']
            v1 = prop[target]
            j1 = result[target]
            v1_type = v1.get('type','')
            if v1_type == 'object':
                ret_value = get_object(v1,j1)
            elif v1_type == 'array':
                ret_value = get_array(v1,j1)
            else:
                ret_value = j1
            return [ret_value]
        
        def get_value(value, schema):
            """
            値を取り出すメソッドです。
            value: jsonから取り出した値です
            schema: imschemaのdataTargetのオブジェクト
            """
            # まずは最優先のdataTargetで指定したEntity typeから求めます
            if schema.get('entityType') != None:   # Entitiesを検索する指定がある
                entityType = schema['entityType']
                if entityType not in run['entitynames']:
                    err.set_errorobj( '01', f'## error ## unknown #attr#, Entity type : {entityType}')
                searchAttr = schema.get('searchAttr','')
                if not searchAttr:
                    err.set_errorobj( '02', f'## schema ## searchAttrの指定が無い. #attr#, entityType: {entityType}')
                valueAttr = schema.get('valueAttr','')
                if valueAttr == '':
                    err.set_errorobj( '03', f'## schema ## valueAttrの指定が無い. #attr#, entityType: {entityType}')
                ent = handles[entityType].get(valueAttr, value)
                if ent== {}:                           # 一致するEntityが無かった
                    handles['trace_log'].set_log(2,'01',f'targetで指定したEntity typeに検索値が一致するEntityが見つかりません. #attr#, entityType: {entityType}, key: {valueAttr}, 検索値: {value}')
                    return ''
                elif searchAttr in ('id', 'type'):
                    ret_value = ent[searchAttr]
                elif searchAttr not in ent:
                    handles['trace_log'].set_log(2,'03',f'valueAttrで検索したEntity内にsearchAttrで指定したAttributeがありません. Entity id: '+ent['id'])
                    return ''
                else:
                    ret_value = ent[searchAttr]['value'] # MUST normalized format
                return ret_value
            else:
                return value
        
        run = self.run
        handles = run['handles']
        err = handles['error']
        ret_list = []
        prop = schema.get('properties',{})
        for key,value in prop.items():    # propertiesのループ
            attrjson = {} if not inputjson else inputjson.get(key,{})
            value_type = value.get('type','')
            if value_type == 'object':
                ret_list += get_object(value,attrjson)
            elif value_type == 'array':        # keyValuesのサポート用にコードを残す
                ret_list += get_array(value,attrjson)
            else:                              # keyValuesのサポート用にコードを残す
                for k1,v1 in value.items():
                    if k1 == 'dataTarget':
                        ret_list += '' if not attrjson else [attrjson] 
                        break
        return ret_list
    
class entity:
    def __init__(self, run):
        self.run=run
        return
    
    def merge (self,
               entities: list,
               entity2: dict,
               schema: dict)->None:
        """
        entityをマージする外部関数
        schema: 実行用スキーマの最上位のdict
        entities:既に生成したEntityの辞書を要素とするリスト
        entity2:マージする側のentityの辞書
        """
        run = self.run
        handles = run['handles']
        if entities == []:
            entities.append(entity2)
            return
        id = entity2['id']
        i1 = len(entities)-1
        find_flag = False
        while i1 >= 0:                         # 性能観点で後ろから探す
            if entities[i1]['id'] == id:
                entity1 = entities[i1]
                find_flag = True
                break
            i1 -= 1
        if find_flag == False:
            entities.append(entity2)
            return
        prop = schema['properties']
        for attr_name, attr_value in prop.items():   # Attribute毎のループ
            if attr_value.get('merge') == None:# merge対象外のAttribute
                continue
            e2 = entity2.get(attr_name,'')
            if e2 == '':                       # mergeする側にAttributeが無い
                continue
            e1 = entity1.get(attr_name,'')
            if e1 == '':                       # mergeされる側にAttributeが無い
                 entity1[attr_name] = e2       # そのまま追加
                 continue
            s1 = attr_value['properties']['value'] # valueの定義内容
            e1 = e1['value']                   # mergeされる側のvalueの値
            e2 = e2['value']                   # mergeする側のvalueの値
            scm_type = s1.get('type')          # Attributeの値のtype
            if scm_type == 'array':
                self.merge_items(e1,e2,s1)
            elif scm_type == 'object':
                self.merge_properties(e1,e2,s1['properties'])
            else:
                self.merge_value(e1,e2,s1)
        return
    
    def merge_items(self, entity1, entity2,schema):
        """
        schemaの記述に従い、リストをマージする。
        schema: schema中のtypeがarrayであるdict
        entity1にentity2をマージする。どちらもリスト
        """
        run = self.run
        handles =run['handles']
        mergekey = schema.get('mergekey','')
        items = schema.get('items','')
        if mergekey == '':
            for e1 in entity2:                 # mergekeyの指定が無いので、単純に連結
                entity1.append(e1)             # 参照渡しにするため、appendで追加
            return                             # リスト全体をマージし終わったので終了
        else:
            if entity2 == []:                  # マージすべき要素がない
                return
            if entity1 == []:
                for e1 in entity2:             # 単純にコピー
                    entity1.append(e1)         # 参照渡しにするため、appendで追加
                return
        
        # ここからは、両方にリストがある場合
        if isinstance(mergekey,str):           # 項目名を指定された
            if len(entity2) != 1:
                handles['error'].set_errorobj('01', f'## error ## サポートしていないentityをマージしようとした')
            obj2 = entity2[0]
            key_value = obj2.get(mergekey,'')  # mergekeyの値を取り出す
            if key_value == '':                # mergekeyの項目が無かった
                handles['error'].set_errorobj('02', f'## error ## mergekeyに指定した項目が無いentityを見つけた')
            k1 = ''
            for obj1 in entity1:               # entity1のリストからキーが一致する要素を探す
                k1 = obj1.get(mergekey,'')
                if k1 == '':
                    continue
                if k1 == key_value:
                    break                          
            if k1 == key_value:                # 一致する項目が見つかった
                scm = items['properties']
                for k1,v1 in scm.items():      # リスト内のsubattributeをサーチ
                    if k1 == mergekey:         # mergekeyはスキップ
                        continue
                    v2 = obj1.get(k1,'')       # entity1側から項目を取り出す
                    if v2 == '':
                        entity1.append(obj2)
                    else:
                        self.merge_properties(obj1,obj2,{k1:v1})
            else:
                entity1.append(entity2[0])
        else:                                  # mergekeyにリストが指定された
            key_attr = mergekey[0]
            key_index = mergekey[1]
            ent2 = entity2[0]
            key_value = ent2[key_attr][key_index]
            f1 = False
            for ent1 in entity1:               # entity1のリストからキーが一致する要素を探す
                if ent1.get(key_attr) != None and ent1[key_attr][key_index] == key_value:
                    f1 = True
                    break
            if f1 == True:                     # 一致するリスト要素が見つかった
                for i2, l2 in enumerate(ent2[key_attr]):
                    if i2 == key_index:
                        continue
                    if i2 >= len(ent1[key_attr]):
                        break
                    for e1 in l2:         # 単純にコピー
                        ent1[key_attr][i2].append(e1)     # 参照渡しにするため、appendで追加
                for ent2_key, ent2_val in entity2[0].items(): # kergekeyと同位のaubattributeをサーチ
                    if ent2_key == key_attr:   # mergekeyの項目はスキップ
                        continue
                    ent1.update(ent2_val)
            else:
                entity1.append(entity2[0])
            return
        
    def merge_properties(self, entity1, entity2, schema):
        """
        entity1: 元のentityのattributeのvalueのdict
        entity2: 追加するentityのAttributeのvalueのdict
        schema: json schemaのtypeがobjectであるdict
        """
        for scm_attr,scm_value in schema.items():  # Attribute毎のループ
            e2 = entity2.get(scm_attr,'')
            if e2 == '':                           # mergeする側にAttributeが無い
                continue
            e1 = entity1.get(scm_attr,'')
            if e1 == '':                           # mergeされる側にAttributeが無い
                entity1[scm_attr] = e2             # そのまま追加
                continue
            t1 = scm_value['type']
            if t1 == 'array':
                self.merge_items(e1,e2,scm_value)
            elif t1 == 'object':
                self.merge_properties(e1,e2,scm_value['properties'])
            else:
                self.merge_value(e1,e2,scm_value)
        return

class trace_log:
    def __init__(self, run):
        self.run = run
        options = run['param']
        err = run['handles']['error']
        if 'logfile' not in options:
            options['logfile'] = ''
        if options['logfile'] == '':
            return
        directory = options['logdir']
        filename = options['logfile']
        if directory != '':
            try:
                os.chdir(directory)
            except Exception as e:
                err.set_errorobj('01', f'## error ## {directory} directory does not exist.')
                raise e
        
        try:
            self.logfile = open(filename, 'w', encoding='utf-8')
        except PermissionError:
            err.set_errorobj('02', f'## error ## {filename} file is already used or not permitted. dir : {directory}')
        except Exception as e:
            err.set_errorobj('03', f'## error ## unknown error occurd. {e}, directory: {directory}, file: {filename}')
        return
    
    def set_log(self, level, id, msg):
        run = self.run
        err = run['handles']['error']
        options = run['param']
        if options['logfile'] == '':
            return
        if level >= 3:
            err.set_errorobj('01', 'invalid level parameter for set_log()')
            return
        attr = run['errorinfo'].get('attr','')
        if attr:
            a = msg.replace('#attr#', f'Attribute: {attr}')
            text = '  '*level + f'{id}: {a}\n'
        else:
            text = '  '*level + f'{id}: {msg}\n'
        n = self.logfile.write(text)    # nは使いません。nは書き込んだ文字数です
        if level >0:
            run['warning'] = True
        return
    
    def __del__(self):
        self.logfile.close()
        return

class error_log:
    def __init__(self, run):
        self.run = run
        return

    def set_errorobj (self, id, msg):
        run = self.run
        caller_frame = inspect.currentframe().f_back
        caller_module = inspect.getmodule(caller_frame)
        caller_class = caller_frame.f_locals.get('__class__')
        caller_function = caller_frame.f_code.co_name
        run['errorflag']=True
        attr = run['errorinfo'].get('attr','')
        a = msg.replace('#attr#', f'Attribute: {attr}') if attr else msg
        run['errorobj'] = {
            'Module': caller_module,
            'Class': caller_class,
            'Function': caller_function,
            'id': id,
            'message': a
        }
        if run['onerror'] == 'exit':
            os.system('pause')
            # exit()
        else:
            print (run['errorobj'])
        os.system('pause')
        sys.exit(1)

if __name__ == '__main__':
    ## テスト用のコードです
    
    ## 変数を初期化します
    run={
        'param':{
            'inputdir': 'c:\\Tools\\data',
            'inputfile': 'Departmentjson.json',
            'outputdir': 'c:\\Tools\\data',
            'outputfile': '',
            'schemadir': 'c:\\Tools\\data',
            'schemafile': 'Departmet.exschema.json',
            'Orgnizationdir': 'c:\\Tools\\data',
            'Orgnizationfile': 'Oganizationjson.json',
            'Departmentdir': 'c:\\Tools\\data',
            'Departmentfile': '',
            'dir': 'c:\\Tools\\data',
            'Landdir': 'c:\\Tools\\data',
            'Landfile': '',
            'Buildingdir': 'c:\\Tools\\data',
            'Buildingfile': 'Buildingjson.json',
            'Facilitydir': 'c:\\Tools\\data',
            'Facilityfile': '',
            'logdir': 'c:\\Tools\\data',
            'logfile': 'jtoclog.txt',
            'jsontype': 'array',
            'truevalue': ['True', 'true', '真', '有'],
            'falsevalue': ['False', 'false', '偽', '無']
        },
        'handles': {},
        'mode': 'debug',
        'errorflag': False,
        'onerror': 'continue',
        'types': ['Text','Number','Integer','Boolean','Date', 'DateTime','URL','Time','Relationship'],
        'entitynames': ['Organization','Department','Land','Building',
            'Facility','FacilityStatistics','BuildingComponent','Device',
            'DeviceModel','Complaint','Report'],
        'errorobj': {},
        'type_tbl': {'str':'string', 'int':'integer','float':'number', 'bool':'boolean' },
        'handles': {}
    }
    run['handles']['error'] = err = error_log(run)
    run['handles']['trace_log'] = trace_log(run)
    run['handles']['schema'] = exschema(run)
    # print (f'checkd schema: ', run['scm'])
    print(json.dumps(run['scm'],ensure_ascii=False, indent=2))
    param = run['param']
    handles = run['handles']
    """
    
    ## 各Entityのファイルを辞書のリストとして読み込みます
    try:
        for entityname in run['entitynames']:
            directory = entityname+'dir'
            filename = entityname+'file'
            if ((filename) in param and param[filename] != ''):
                handles[entityname] = GET_entities(run,entityname)
                handles[entityname].load(param[entityname + 'dir'],param[entityname + 'file'])
                handles['trace_log'].set_log(0,'04',f'directory: {directory}, {entityname}file: {filename}')
                if not isinstance(handles[entityname].entities,list):
                    err.set_errorobj('073', f'## error ## Entitiesの形式がリスト構造ではありません. Entity type: {entityname}')
                if len(handles[entityname].entities) == 0:
                    err.set_errorobj('08', f'## error ## Entitiesの件数がゼロ件です. Entity type: {entityname}')
                if handles[entityname].entities[0].get('id') == None:
                    err.set_errorobj('09', f'## error ## Entityの:形式が不正です. Entity type: {entityname}')
    except FileNotFoundError as e:
        handles['error'].set_errorobj('10', f'## error ## Entities load  error : {e}\n')
    handles['trace_log'].set_log(0,'09','outputdir: '+param['outputdir']+', outputfile: '+param['outputfile'])
    
    ## 拡張JSONスキーマを辞書として読み込みます
    handles['schema'] = exschema(run)
    
    ## 入力のJSON文(EntityのArray)を読み込みます
    filename = param['inputfile']
    directory = param['inputdir']
    handles['inputjson'] = GET_entities(run,'inputjson')
    handles['inputjson'].load(directory,filename)
    in_list = handles['inputjson'].entities
    handles['trace_log'].set_log(0,'04',f'directory: {directory}, inputfile: {filename}')
    if not isinstance(handles['inputjson'].entities,list):
        err.set_errorobj('05', f'## error ## inputfileのEntitiesの形式がリスト構造ではありません.')
    if len(handles['inputjson'].entities) == 0:
        err.set_errorobj('06', f'## error ## inputfileのEntitiesの件数がゼロ件です.')
    if handles['inputjson'].entities[0].get('id') == None:
        err.set_errorobj('07', f'## error ## inputfileのEntityの:形式が不正です.')
    header = run['handles']['schema'].get_header(run['scm'])
    for in_entity in in_list:
        rr = run['handles']['schema'].get_row(run['scm'], in_entity)
    """