#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# checked with 3.10,3.11
# Created by I.Watanabe on 2023-12-25.
# Customized for NGSI v2 by H. Inoue on 2024-02-10 

import csv
import os
import re
from my_package.kanji_converter import KanjiConverter, K_DIGIT, PAT_DIGIT

class address:
    @classmethod
    def convert2dict(cls, address_list):
        address_dict = {}
        for address in address_list:
            prefecture = address[1]
            county = address[4]
            city = address[7]
            ward = address[10]
            
            if prefecture not in address_dict:
                address_dict[prefecture] = {}
            
            if county:
                if county not in address_dict[prefecture]:
                    address_dict[prefecture][county] = {}
                if city not in address_dict[prefecture][county]:
                    address_dict[prefecture][county][city] = {}
            elif city not in address_dict[prefecture]:
                address_dict[prefecture][city] = {}
            
            if ward and ward not in address_dict[prefecture][city]:
                address_dict[prefecture][city][ward] = {}
        return address_dict
    
    @classmethod
    def convert2index(cls, address_list):
        address_index = {}
        for address in address_list:
            prefecture = address[1]
            county = address[4]
            city = address[7]
            ward = address[10]
            if county in ('色丹郡', '国後郡', '択捉郡', '紗那郡', '蘂取郡'):
                continue                   # 北方領土があると村名が重複
            if prefecture not in address_index:
                address_index[prefecture] = {
                    'start': 'pref',
                    'pref':prefecture
                }
            if county:
                if county+city not in address_index:
                    address_index[county+city] = {
                        'start': 'county',
                        'pref':prefecture,
                        'county': county,
                        'city': city,
                        'ward': ward
                    }
                else:
                    address_index[county] = {'duplicate': True}
            if city not in address_index:
                address_index[city] = {
                    'start': 'city',
                    'pref':prefecture,
                    'county': county,
                    'city': city,
                    'ward': ward
                }
            elif not ward:
                address_index[city] = {'start': 'dupname'}
        return address_index
    
    @classmethod
    def convertStreetNumbers(cls, address_str: str) -> str:
        """入力文字列から丁目番地を削除し'-'で結合した数字の組に変える
        - 最終端が[号、番、番地、丁目]で終わっている場合は'-'は削除する
        - 丁目番地が含まれていない文字列はそのままとする
        - 文字列の最後が数字で終わっていない場合は変換を行わない
        例：
        1丁目3番地10号hoge -> 1丁目3番地10号hoge  # 最終文字が数字でないため変換しない
        〇〇2034 -> 〇〇2034
        〇〇 -> 〇〇
        1丁目3番地10号123 -> 1-3-10-123
        1丁目3番地10号 -> 1-3-10
        1丁目3番地 -> 1-3
        1丁目3番 -> 1-3
        1丁目10号 -> 1-10
        1丁目 -> 1
        3番地10号 -> 3-10
        3番地 -> 3
        3番 -> 3
        10号 -> 10
        """
        # '-'に置き換える文字
        PAT_DELETE = rf"(?<=\d)(丁目|番地|番|号室?|の|ノ)[のノ]?"
        
        if not address_str:
            return address_str, '', None  # 空文字なら何もしない
        
        # 半角カナの全角化と全角数字の半角化
        normalized = KanjiConverter.normalizeStr(address_str)
        # 都道府県名の補完
        # normalized = cls.fill_prefecture(normalized)
        # 丁目番地の数字をを半角に変換
        normalized = cls.convNumInStreet(normalized)
        
        # 丁目番地の文字を'-'に置換
        replaced = cls.replace_number_blocks(normalized, PAT_DELETE, '-')
        replaced = re.sub(r'-$', "", replaced)  # 最後の'-''を消す
        # # 最後の数字ブロックのみ取り出して一緒に返却
        return replaced
    
    @classmethod
    def convNumInStreet(cls, str_arg: str, pat: str = None) -> str:
        """漢数字の半角数字変換
                丁目番地表記部分の漢数字を半角数字にする。
                千代田区などの番地表記以外の漢数字は変換しない
        """
        # 漢数字変換対象の丁目番地の後置パターン
        PAT_ENDING = rf"(丁目|番地|号室?|[のノ](?={PAT_DIGIT}))(?=[のノ]*)"
        # 変換対象となる丁目番地の漢数字パターン
        TargetDefault = rf'[{K_DIGIT}]+(?={PAT_ENDING})|[{K_DIGIT}]+$'
        Taget = pat if pat else TargetDefault
        matches = list(re.finditer(Taget, str_arg))
        for match in reversed(matches):
            kanji_num = match.group()
            start, end = match.span()
            ascii_num = KanjiConverter.kanji2ascii(kanji_num)
            str_arg = str_arg[:start] + ascii_num + str_arg[end:]
        return str_arg
    
    @classmethod
    def replace_number_blocks(cls, text, pattern, replacement):
        """Helper method to replace number blocks with given replacement character.
        """
        matches = list(re.finditer(pattern, text))
        # 後ろから置き換える
        for match in reversed(matches):
            start, end = match.span()  # 対象位置を取得
            text = text[:start] + replacement + text[end:]
        return text
    
    @classmethod
    def search_in_address(cls, text: str, parts: dict):
        """住所の要素部分を探して特定するヘルパー関数"""
        for key, value  in parts.items():
            if text.startswith(key):
                return key, value
        return '', ''
    
    @classmethod
    def parse(cls, run,text: str, address: dict) -> None:
        # Prefectureの取得
        address_table = run['address_table']
        address_index = run['address_index']
        trace = run['handles']['trace_log']
        key, value = cls.search_in_address(text, address_index)
        if not key:                        # 市内の住所から始まっている
            pass
        elif value['start'] == 'pref': # 県名から始まっている
            pref=key
            text = text[len(key):]
            key, value = cls.search_in_address(text, address_table[pref])
            if not key:
                trace.set_log(2,'01',f'## error ## 都道府県名内に郡名または市区町村名が存在しない. address: {text}')
                return
            if key[-1:] == '郡':
                county = key
                text = text[len(county):]
                town, value = cls.search_in_address(text, address_table[pref][county])
                print(f'county: {county}, town: {town}')
                address['addressLocality'] = county + town
                text = text[len(town):]
            else:
                address['addressLocality'] = key
                text = text[len(key):]
        elif value['start'] == 'county':   # 郡名+町村名から始まっている
            address['addressRegion'] = value['pref']
            address['addressLocality'] = key
            text = text[len(key):]
        elif value['start'] == 'city':     # 市町村名か特別区名から始まっている
            address['addressRegion'] = value['pref']
            address['addressLocality'] = value['county']+value['city']
            text = text[len(key):]
        else:                              # 名称が重複する自治体名から始まっている
            if not address['addressRegion']:
                trace.set_log(2,'02',f'## error##: 重複する自治体名では、都道府県名の既定値は必須です. address: {text}')
                return
            pref = address['addressRegion']
            city = ''
            for key, value in address_table[pref].items():
                if key[-1:] == '郡':
                    print (f'郡のループ: 郡: {key}, 自治体',value)
                    for k2 in value:
                        if k2 == text[:len(k2)-1]:
                            city = key+k2
                            break
                elif key == text[:len(key)]:
                    county = ''
                    city = key
                    break
                if city:
                    break
            address['addressLocality'] = city
            text = text[len(city):]
        # アドレスの正規化と丁目番地変換
        text = cls.convertStreetNumbers(text)
        address['streetAddress'] = text
        return

def load_csv(run, directory, filename):
    handles = run['handles']
    try:
        os.chdir(directory)
    except FileNotFoundError as e:
        handles['error'].set_errorobj('01', f'## error ## {directory} directory does not exist.')
    
    try:
        with open(filename, 'r', encoding='utf-8') as f:
            try:
                reader = csv.reader(f)
                tbl = [row for row in reader]
            except csv.Error as e:
                handles['error'].set_errorobj('02', f'CSV error : {e}')
    except FileNotFoundError as e:
        handles['error'].set_errorobj('02', f'## error ## {filename} does not exist. dir: {directory}, file: {filename}')
    if tbl and tbl[0][0].startswith('\ufeff'):
        tbl[0][0] = tbl[0][0][1:]
    return tbl
