#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# checked with 3.11
# Created by I.Watanabe on 2023-12-25.

"""
このスクリプトは、指定された文字列内の漢数字をアラビア数字に変換します。
与えられた入力が一部またはすべて漢数字で構成されている場合、ここに定義された関数を使用してこれをアラビア数字に変換できます。
また、文字列を正規化するための関数も提供しています。

使用例:
>>> KanjiConverter.kanji2ascii('千代田区霞が関二丁目1-6')
'千代田区霞が関2-1-6'
"""
import jaconv
from kanjize import kanji2number

# 漢数字と見做す文字
K_DIGIT = '〇零一二三四五六七八九十百千万'
# 漢数字と半角数
PAT_DIGIT = rf'[{K_DIGIT}\d]'


class KanjiConverter:

    @classmethod
    def normalizeStr(cls, atr_arg: str) -> str:
        """全角英数記号を半角に、半角カナを全角に変換する関数
        Args:
            atr_arg (str): 変換対象の文字列.
        Returns:
            str: 正規化された文字列.
        """
        if not atr_arg:
            return atr_arg

        half_alnum = jaconv.z2h(atr_arg, ascii=True, digit=True)
        full_kana = jaconv.h2z(half_alnum, kana=True)
        return full_kana

    @classmethod
    def k2digit(cls, num: str) -> int:
        """漢数字とアラビア数字が混在した文字列を整数に変換する
        Args:
            num (str): アラビア数字と漢数字が混在した文字列.
        Returns:
            int: 変換後の整数値.
        Raises:
            ValueError: 入力が整数値に変換できない場合に発生する例外.
        """
        return int(kanji2number(num))

    @classmethod
    def kanji2ascii(cls, str_arg: str) -> str:
        """
        漢字の数字をASCIIに変換する。アラビア数字に変換できない場合は、元の文字列をそのまま返します。
        Args:
            str_arg (str): 変換対象の文字列.
        Returns:
            str: 漢字の数字がアラビア数字に変換された文字列、もしくは元の文字列。
        """
        try:
            text_ = cls.k2digit(str_arg)
        except ValueError:
            return str_arg

        return str(text_)


if __name__ == '__main__':
    DATA = [
        '博多駅中央街八番街1-1',
        '梅田三番街3丁目5番3号一一〇九',
        '福岡県福岡市博多区博多駅中央街八番街1-1',
        '大阪府大阪市北区梅田三番街3丁目5番3号一〇九',
        # '北海道札幌市北区北二十四条西２０丁目8',
        # '北海道札幌市北区北二十四条西２０丁目',
        # '北海道札幌市北区北二十四条西十五',
        '北海道札幌市北区北二十四条西三十六丁目三〇四号',
        '北海道札幌市北区北二十四条西10-5-6-1',
        '北海道札幌市北区北二十四条西10-5-6',
        '北海道札幌市北区北二十四条西10-5',
        '東京都武蔵村山市本町一丁目1番地の1',
        '東京都武蔵村山市本町一丁目1ノ1',

        '千代田区霞が関2-1-6マンション千代田三1102',
        '千代田区霞が関2-1-6マンション千代田九 一〇二',
        '千代田区霞が関2-1-6マンション千代田三一〇二',
        '千代田区霞が関2-1-6マンション千代田三零〇四',
        '千代田区霞が関2-1-6マンション千代田二十1の三の一〇二',
        '千代田区霞が関2-1-6 一三〇一',
        '千代田区霞が関2-1-6マンション千代田三1の3 一〇二',
        '札幌市北区北二十四条西六丁目',
        ]
    x = '梅田三番街83-5番地4番の3号7ノ33-1109'
    y = '愛知県名古屋市中区三の丸三丁目1番2号'
    PAT = rf"(?<=\d)(丁目|番地|番|号|[のノ])[のノ]?"
