#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# checked with 3.10
# Created by H. Inoue on 2024-02-10 
"""
Coppell Technologiesの各種ツールから利用する共通関数群です
本クラス群は汎用的に利用される事を想定しています。
基本的に例外は処理していないため、呼び出し側でハンドリングが必要です
"""
import os
import json
import csv
import re
## Coppell Technologiesが提供するツール群の共通クラスです

class CT_json:
    ## json関連共通クラスです
    
    ## ファイル内のJSON文をdictに展開します。返却値は展開したdictです
    @classmethod
    def load (cls, directory, filename):
        if directory != '':
            os.chdir(directory)
        with open(filename, 'r', encoding='utf-8') as f:
            return json.loads(f.read())
    
    @classmethod
    def save (cls, directory, filename, data, mode='w'):
        if directory != '':
            os.chdir(directory)
        with open(filename, mode, encoding='utf-8') as f:
            return json.dump(data, f,indent=2, ensure_ascii=False)

class CT_csv:
    ## csv関連共通クラスです
    
    ## ファイル内のcsvを二次元のlistに展開します。返却値は展開したlistです
    @classmethod
    def load(cls, directory, filename):
        os.chdir(directory)
        with open(filename, 'r', encoding='utf-8') as f:
            reader = csv.reader(f)
            tbl = [row for row in reader]
        # Excelでcsvを作成した場合に付加されるBOMです
        if tbl and tbl[0][0].startswith('\ufeff'):
            tbl[0][0] = tbl[0][0][1:]
        return tbl
    
    @classmethod
    def save(cls, directory, filename, data, mode='w'):
        if directory != '':
            os.chdir(directory)
        with open(filename, mode, newline='', encoding='utf-8') as f:
            writer = csv.writer(f)
            writer.writerows(data)

class CT_text:
    ## text関連共通クラスです
    
    ## ファイル内のtextを出力します。モードの既定値はwriteです
    @classmethod
    def save(cls, directory, filename, text, mode='w'):
    ## テキストをファイルとして出力します
        if directory != '':
            os.chdir(directory)
        with open(filename, mode, encoding='utf-8') as f:
            return f.write(text)                          # nは使いません。nは書き込んだ文字数です
    
    ## 文字列が数値に変換可能が確認します
    @classmethod
    def isnumeric(cls, text)->bool:
    ## テキストを数値に変換可能か調べます
        return re.match(r'^\d+(\.\d+)?$', text) is not None
    
    ## 文字列が数値に変換可能が確認します
    @classmethod
    def float(cls, text)->float:
    ## テキストを数値に変換して返却します
        if '.' in text:
            return float(text)
        else:
            return int(text)

class CT_list:
    ## listを操作する共通クラスです
    def flatten_nested_list(in_list: list)->list:
        """
        2重listを展開します。listは外側も内側も順序無しと見做します。
        処理結果は、内側の順序無しの要素が一つになります。3重以上のlist
        の場合は、外側から2つのlistが処理対象となります。内側のlist
        が複数あり、ネスト関係に無い場合は、最初に見つけたlistが処理
        対象となります
        in_list = [{x:b,y:{z:[c,[d,e,f]]}}]の場合、返却値は、以下となります。
        [{x:b,y:{z:[c]},{x:b,y:{z:[[d,e,f]]}}]
        """
        def flatten_nested_list_sub(in_dict)->list:
            """
            内側のlistを自己回帰的に探してlistが見つかったらlistの要素を
            一つずつ分解して、dictの列として返却します
            """
            ret_list = []
            for outer_key, outer_value in in_dict.items():
                if isinstance(outer_value,dict):
                    inner_list = flatten_nested_list_sub(outer_value)
                    for inner_item in inner_list:
                        in_dict_copy = in_dict.copy()
                        in_dict_copy[outer_key] = inner_item
                        ret_list += [in_dict_copy]
                elif isinstance(outer_value,list):
                    # 内側のリストを一つずつの要素に分解する
                    for inner_item in outer_value:
                        ret_list += [{outer_key:[inner_item]}]
                else:
                    # 外側のlistの要素が値だった
                    ret_list += [{outer_key:outer_value}]
            return ret_list
        
        # メソッドの本体です
        ret_list = []
        flags = {'found':False}
        for outer_item in in_list:
            if isinstance(outer_item, dict):
                # 外側のlistの要素がdictだった
                inner_list = flatten_nested_list_sub(outer_item)
                ret_list += inner_list
            elif isinstance(outer_item, list):
                # 外側のlistの要素がlistだった
                flags['found'] = True
                for inner_item in outer_item:
                    ret_list += [[inner_item]]
            else:
                # 外側のlistの要素が値だった
                ret_list += [outer_item]
        return ret_list
    
    def flatten_ordered_list(in_list)->list:
        """    
        3重リストを展開します。リストは外側から順序無し、順序付き、順序無しの
        構造と見做します。従って、一番内側の順序無しの要素が一つになり、一番外側
        の順序無しリストに展開します。4重以上のリスト
        の場合は、外側から3つのリストが処理対象となります。
        in_list = [{x:b,y:{z:[c,[d,e,f]]}]の場合、返却値は、以下となります。
        [{x:b,y:{z:[c,[d]]},{x:b,y:{z:[c,[e]]},{x:b,y:{z:[c,[f]]}]
        """
        def flatten_ordered_list_dict(in_dict)->list:
            """
            dictの中にlistがある構造です。最初に見つけたlistの要素を一つずつに分解し、dictのlistに変換します、
            flags['ordered']がTrueの時は最初のlistをFalseの場合は二つ目のlistを分解します。
            例えば
            in_dict = {a: [{b:c},{d:[e,f]}]}の場合、Trueであれば、
            [{a: [{b:c}]},{a: [{d:[e,f]}]}]を返却し、Falseであれば、
            [{a: [{b:c},{d:[e]}]},{a: [{b:c},{d:[f]}]}]を返却します
            """
            print (f'dict. in_dict', in_dict)
            a =10/0
            ret_list = []
            for outer_key, outer_value in in_dict.items():
                if isinstance(outer_value,dict):
                    inner_list = flatten_ordered_list_dict(outer_value)
                    print(f'dict(dict.1). inner_list', inner_list, f'outer_value', outer_value)
                    for inner_item in inner_list:
                        ret_list += [{outer_key:inner_item}]
                    print(f'dict(dict.inner). new in_dict.item', {outer_key:inner_item})
                elif flags['ordered'] and isinstance(outer_value,list):
                    # 内側のリストを一つずつの要素に分解する
                    for inner_item in outer_value:
                        ret_list += [{outer_key:[inner_item]}]
                elif not flags['ordered'] and isinstance(outer_value,list):
                    # 内側のリストを一つずつの要素に分解する
                    inner_list = flatten_ordered_list_list(outer_value)
                    for inner_item in inner_list:
                        ret_list += [{outer_key:inner_item}]
                    print(f'dict(list). ret_list', ret_list)
                else:
                    # 外側のlistの要素が値だった
                    ret_list += [{outer_key:outer_value}]
            print(f'dict(return). ret_list', ret_list)
            return ret_list
        
        def flatten_ordered_list_list(in_list)->list:
            """
            3重のlistの内、2番目のordered listの処理
            in_list = [{'a':['b','c']},{'d':['e','f']}]であれば、['b','c']だけを分解し
            返却値は [[{'a':['b']},{'d':['e','f']}],[{'a':['c']},{'d':['e','f']}]]
            in_list = ['a',['b','c','d']であれば、
            返却値は [['a':['b']],['a':['c']],['a':['d']]]
            """
            print (f'list. in_list', in_list)
            flags['ordered'] = True
            ret_list = []
            for index, ordered_item in enumerate(in_list):
                if isinstance(ordered_item,dict):# ordered listの要素はdictだった
                    inner_list = flatten_ordered_list_dict(ordered_item)
                    for inner_item in inner_list:
                        in_list_copy = in_list.copy()
                        in_list_copy[index] = [inner_item]
                        ret_list += [in_list_copy]
                    print(f'list(dict.return). ret_list', ret_list)
                    ret_list += inner_list
                elif isinstance(ordered_item, list):
                    # 1番内側のlistが見つかった
                    flags['inner'] = True
                    for inner_item in ordered_item:
                        in_list_copy = in_list.copy()
                        in_list_copy[index] = [inner_item]
                        ret_list += [in_list_copy]
                    print(f'list(list.return). ret_list', ret_list)
                    break
            if not ret_list:
                ret_list += [in_list]
            return ret_list
        
        # メソッドの本体です
        ret_list = []
        # protected: あるdict内にlistが見つかったので、他のdict内のlistは無視する
        # found_innter: 内側のlistが既に見つかった
        if len(in_list) == 0:                  # 入力のlistが空だった
            return []
        for outer_item in in_list:
            flags ={'ordered':False}
            if isinstance(outer_item, dict):   # 外側のlistの要素はdictだった
                # ordered listはまだ見つかっていない
                middle_list = flatten_ordered_list_dict(outer_item)
                ret_list += middle_list
                print(f'main(dict). ret_list', ret_list)
            elif isinstance(outer_item, list):
                # ordered listが見つかった
                inner_list = flatten_ordered_list_list(outer_item)
                ret_list += inner_list
            else:                                  # 外側のlistの要素が値だった
                ret_list += [outer_item]
        return ret_list                 # ネスト構造が見つからなかった
    
# in_list = ['a','b','c']
# in_list = [['a','b','c'],['d','e','f']]
# in_list = [{'x':['a','b','c']},{'y':['d','e','f']}]
# in_list = [{'x':{'zx':['a','b','c']}},{'y':{'zy':['d','e','f']}}]
# in_list = [{'x':{'zx':{'zz':['a','b','c']}}},{'y':{'zy':{'zz':['d','e','f']}}}]
# in_list = [{'g':'h'},{'x':{'zx':{'zz':['a','b',['c1','c2']]}}},{'y':{'zy':{'zz':['d',['e1','e2'],'f']}}}]
# in_list = [{'a': {'g': {'h': 'b'}}, 'e': 'f'}]
# in_list = [[{'a': {'g': {'h': 'b'}}}, {'e': 'f'}]]
# in_list = [{'x':'b','y':{'z':['c',['d','e','f']]}}]
# in_list = [{'a': {'g': {'h': ['b', 'c', 'd']}}, 'e': ['f']}]
# in_list = [{'a': {'g': {'h': ['i',['b', 'c', ['d','j']]]}}, 'e': ['f']}]
# in_list = [{'a': {'g': {'h': ['i',['b', 'c', 'd']]}}, 'e': ['f']}]
# in_list = [
#    {"subzoneName": ["1F",["吉浦支所","男子トイレ","女子トイレ","正面玄関","階段室","共用ロビー"]]},
#    {"subzoneName": ["2F",["市民センター窓口","音楽室","調理室","図書室","階段室","共用ロビー"]]},
#    {"subzoneName": ["3F",["学習室1","学習室2","階段室","共用ロビー"]]},
#    {"subzoneName": ["4F",["ホール","控室1","階段室","共用ロビー"]]}
#]
# flattened_list = CT_list.flatten_nested_list(in_list)
#flattened_list = CT_list.flatten_ordered_list(in_list)
#print('in_list       : ',in_list,'\nflattened_list: ', flattened_list)
